# -*- coding: utf-8 -*-

import pytest
import threading
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import CallbackRegistry, SignalWatcher, SampleTasks

from pyknic.lib.tasks.proto import SchedulerProto, SchedulerFeedback, ScheduleSourceProto, ScheduledTaskPostponePolicy
from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.threaded_task import ThreadedTask
from pyknic.lib.tasks.scheduler.plain_sources import InstantTaskSource
from pyknic.lib.tasks.scheduler.scheduler import Scheduler
from pyknic.lib.tasks.scheduler.record import ScheduleRecord


class TestScheduler:

    class SynCallback:
        # TODO: refactor this

        def __init__(self) -> None:
            self.event = threading.Event()

        def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
            self.event.set()

        @staticmethod
        def flush(source: InstantTaskSource) -> None:
            syn = TestScheduler.SynCallback()
            source.schedule_record(ScheduleRecord(PlainTask(syn), source))
            syn.event.wait()

    def test(self, callbacks_registry: 'CallbackRegistry') -> None:
        scheduler = Scheduler()
        source1 = InstantTaskSource()
        source2 = InstantTaskSource()

        threaded_scheduler = ThreadedTask(scheduler)
        threaded_scheduler.start()

        scheduler.subscribe(source1)
        scheduler.subscribe(source2)

        source1.schedule_record(ScheduleRecord(PlainTask(callbacks_registry.callback('test-callback')), source1))
        TestScheduler.SynCallback.flush(source2)
        assert(callbacks_registry.calls('test-callback') == 1)

        scheduler.unsubscribe(source1)
        source1.schedule_record(ScheduleRecord(PlainTask(callbacks_registry.callback('test-callback')), source1))
        TestScheduler.SynCallback.flush(source2)
        assert(callbacks_registry.calls('test-callback') == 1)

        threaded_scheduler.stop()
        threaded_scheduler.join()

    def test_exception(self) -> None:
        scheduler = Scheduler()
        source = InstantTaskSource()

        threaded_scheduler = ThreadedTask(scheduler)
        threaded_scheduler.start()

        with pytest.raises(ValueError):
            scheduler.unsubscribe(source)

        scheduler.subscribe(source)

        with pytest.raises(ValueError):
            scheduler.subscribe(source)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_scheduled(self, signal_watcher: 'SignalWatcher') -> None:
        scheduler = Scheduler()
        source = InstantTaskSource()
        record = ScheduleRecord(PlainTask(lambda: None), source)

        scheduler.callback(SchedulerProto.task_scheduled, signal_watcher)

        threaded_scheduler = ThreadedTask(scheduler)
        threaded_scheduler.start()

        scheduler.subscribe(source)
        source.schedule_record(record)

        signal_watcher.wait(10)  # will raise an exception in case of failure

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_feedback(self, callbacks_registry: 'CallbackRegistry') -> None:

        class CustomSource(ScheduleSourceProto):

            def scheduler_feedback(self, scheduler: 'SchedulerProto', feedback: SchedulerFeedback) -> None:
                if feedback == SchedulerFeedback.source_subscribed:
                    callbacks_registry.callback('source-subscribed')()
                elif feedback == SchedulerFeedback.source_unsubscribed:
                    callbacks_registry.callback('source-unsubscribed')()

        source = CustomSource()
        scheduler = Scheduler()
        threaded_scheduler = ThreadedTask(scheduler)
        threaded_scheduler.start()
        assert(callbacks_registry.calls("source-subscribed") == 0)
        assert(callbacks_registry.calls("source-unsubscribed") == 0)

        scheduler.subscribe(source)
        assert(callbacks_registry.calls("source-subscribed") == 1)
        assert(callbacks_registry.calls("source-unsubscribed") == 0)

        scheduler.unsubscribe(source)
        assert(callbacks_registry.calls("source-subscribed") == 1)
        assert(callbacks_registry.calls("source-unsubscribed") == 1)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_dropped(
        self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks'
    ) -> None:
        source = InstantTaskSource()
        scheduler = Scheduler(1)
        threaded_scheduler = ThreadedTask(scheduler)

        scheduler.callback(SchedulerProto.scheduled_task_dropped, signal_watcher)
        threaded_scheduler.start()
        scheduler.subscribe(source)

        source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(terminate_method=False), source))
        source.schedule_record(ScheduleRecord(
            sample_tasks.LongRunningTask(terminate_method=False),
            source,
            postpone_policy=ScheduledTaskPostponePolicy.drop
        ))
        signal_watcher.wait(100)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_postponed(self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks') -> None:
        source = InstantTaskSource()
        scheduler = Scheduler(1)
        threaded_scheduler = ThreadedTask(scheduler)

        scheduler.callback(SchedulerProto.scheduled_task_postponed, signal_watcher)
        threaded_scheduler.start()
        scheduler.subscribe(source)

        source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(terminate_method=False), source))
        source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(terminate_method=False), source))
        signal_watcher.wait(100)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_expired(self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks') -> None:
        source = InstantTaskSource()
        scheduler = Scheduler(1)
        threaded_scheduler = ThreadedTask(scheduler)

        scheduler.callback(SchedulerProto.scheduled_task_expired, signal_watcher)
        threaded_scheduler.start()
        scheduler.subscribe(source)

        source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(terminate_method=False), source, ttl=1))
        signal_watcher.wait(100)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_started(self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks') -> None:
        source = InstantTaskSource()
        scheduler = Scheduler(1)
        threaded_scheduler = ThreadedTask(scheduler)

        scheduler.callback(SchedulerProto.scheduled_task_started, signal_watcher)
        threaded_scheduler.start()
        scheduler.subscribe(source)

        source.schedule_record(ScheduleRecord(sample_tasks.LongRunningTask(terminate_method=False), source))
        signal_watcher.wait(100)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()

    def test_signal_task_complete(self, signal_watcher: 'SignalWatcher', sample_tasks: 'SampleTasks') -> None:

        source = InstantTaskSource()
        scheduler = Scheduler(1)
        threaded_scheduler = ThreadedTask(scheduler)

        scheduler.callback(SchedulerProto.scheduled_task_completed, signal_watcher)
        threaded_scheduler.start()
        scheduler.subscribe(source)

        source.schedule_record(ScheduleRecord(sample_tasks.DummyTask(), source))
        signal_watcher.wait(100)

        threaded_scheduler.stop()
        threaded_scheduler.wait()
        threaded_scheduler.join()



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/models/tg_bot_types.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: autogenerate like this "https://github.com/AntonOvsyannikov/pure-teleapi/blob/main/apigen/__main__.py"

import pydantic
import typing


# These models are based on https://core.telegram.org/bots/api#available-types


class CallbackQuery(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#callbackquery
    id_: str = pydantic.Field(alias='id')
    from_: 'User' = pydantic.Field(alias='from')
    data: str | None = None

    model_config = pydantic.ConfigDict(
        populate_by_name=True,
    )


class Chat(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#chat
    id_: int = pydantic.Field(alias='id')
    type_: str = pydantic.Field(alias='type')  # TODO: better to do it with enum

    model_config = pydantic.ConfigDict(
        populate_by_name=True,
    )


class InlineKeyboardButton(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#inlinekeyboardbutton
    text: str
    callback_data: str | None = None


class InlineKeyboardMarkup(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#inlinekeyboardmarkup
    inline_keyboard: typing.List[typing.List[InlineKeyboardButton]]


class Message(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#message
    message_id: int
    from_: typing.Optional['User'] = pydantic.Field(default=None, alias='from')
    chat: Chat
    text: typing.Optional[str] = None

    model_config = pydantic.ConfigDict(
        populate_by_name=True,
    )


class ReplyParameters(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#replyparameters
    message_id: int


class Update(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#update
    update_id: int
    callback_query: CallbackQuery | None = None
    message: Message | None = None


class User(pydantic.BaseModel):
    # origin: https://core.telegram.org/bots/api#user
    id_: int = pydantic.Field(alias='id')
    language_code: str | None = None

    model_config = pydantic.ConfigDict(
        populate_by_name=True,
    )



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/virtual_dir.py
#
# Copyright (C) 2018-2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import functools
import os.path
import pathlib
import typing

from pyknic.lib.verify import verify_value
from pyknic.lib.uri import URI
from pyknic.lib.io.clients.proto import IOClientProto
from pyknic.lib.io.aio_wrapper import AsyncWrapper


@verify_value(path=lambda x: x.is_absolute())
def path_to_str(path: pathlib.PosixPath, relative_path: bool = False) -> str:
    # TODO: test it and document!
    if not relative_path:
        return os.path.abspath(path)  # TODO: Check whether it will work on Windows?!

    return os.path.abspath(path).lstrip('/')


class VirtualDirectoryClient(IOClientProto):
    """This class may be used as a basic class for network clients, that does not keep connection between
    capabilities calls. That type of clients may create connection for each request. In this case it is important
    to save current session directory, because it influences calls behaviour. This class helps to save and
    to change current session directory during such calls."""

    @classmethod
    def create_client(cls, uri: URI) -> 'VirtualDirectoryClient':
        """Basic client creation."""
        return cls(uri)  # type: ignore[no-any-return]  # mypy issue

    @verify_value(start_path=lambda x: x is None or x.is_absolute())
    def __init__(
        self,
        uri: URI,
        start_path: typing.Optional[pathlib.PosixPath] = None
    ):
        """Create a client

        :param uri: URI to connect to.
        :param start_path: Optional path to start with. If not defined then '/' is used
        """
        IOClientProto.__init__(self)

        self.__uri = uri
        self.__session_path = start_path if start_path else pathlib.PosixPath('/')

    def uri(self) -> URI:
        """Create URI this client was constructed with."""
        return self.__uri

    def join_path(self, *path: str) -> pathlib.PosixPath:
        """Append current path with entries (directories) and return current path.

        :param path: path items to join
        """
        for p in path:
            self.__session_path /= p.lstrip('/')

        return self.session_path()

    @verify_value(path=lambda x: x is None or x.is_absolute())
    def session_path(self, path: typing.Optional[pathlib.PosixPath] = None) -> pathlib.PosixPath:
        """Set and/or get current session path.
        :param path: If defined then this path will be used as a current session path
        """
        if path is not None:
            self.__session_path = path

        return self.__session_path

    @verify_value(entry=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def entry_path(self, entry: str) -> pathlib.PosixPath:
        """Return path of a file that is stored inside a current session path.

        :param entry: inner file/directory name
        """
        return self.__session_path / entry

    def _wrap_capability(self, implementation: object, method_name: str) -> None:
        async def wrapper_fn(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            method_fn = getattr(implementation, method_name)
            caller = await AsyncWrapper.create(functools.partial(method_fn, *args, **kwargs))
            return await caller()

        self.append_capability(getattr(IOClientProto, method_name), wrapper_fn)



# -*- coding: utf-8 -*-

import threading
import typing
import pytest

from pyknic.lib.signals.proto import SignalSourceProto, Signal


class CallbackRegistry:

    def __init__(self) -> None:
        self.__lock: typing.Optional[threading.Lock] = None
        self.__calls: typing.Dict[typing.Hashable, int] = dict()

    def thread_safe(self) -> None:
        assert(not self.__lock)
        self.__lock = threading.Lock()

    def callback(self, callback_id: typing.Hashable = None, callback_result: typing.Any = None) -> 'CallbackCall':
        return CallbackCall(self, callback_id, callback_result)

    def calls(self, callback_id: typing.Hashable = None) -> int:
        return self.__calls.get(callback_id, 0)

    def total_calls(self) -> int:
        return sum(self.__calls.values())

    def register_call(self, callback_id: typing.Hashable = None) -> None:
        def increase() -> None:
            cnt = self.__calls.setdefault(callback_id, 0)
            self.__calls[callback_id] = (cnt + 1)

        if self.__lock:
            with self.__lock:
                increase()
        else:
            increase()


class CallbackCall:

    def __init__(self, registry: CallbackRegistry, callback_id: typing.Hashable, result: typing.Any = None):
        self.__registry = registry
        self.__callback_id = callback_id
        self.__result = result

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        self.__registry.register_call(self.__callback_id)
        if callable(self.__result):
            return self.__result()
        return self.__result


SignalType: typing.TypeAlias = typing.Tuple[
    typing.Optional[SignalSourceProto],
    typing.Optional[Signal],
    typing.Optional[typing.Any]
]


class SignalsRegistry:

    def __init__(self) -> None:
        self.__lock: typing.Optional[threading.Lock] = None
        self.__calls: typing.List[SignalType] = list()

    def thread_safe(self) -> None:
        assert(not self.__lock)
        self.__lock = threading.Lock()

    def flush(self) -> None:
        self.__calls.clear()

    def __call__(self, source: SignalSourceProto, signal: Signal, signal_value: typing.Any) -> None:
        if not self.__lock:
            self.__calls.append((source, signal, signal_value))
        else:
            with self.__lock:
                self.__calls.append((source, signal, signal_value))

    def dump(self, flush: bool = False) -> typing.List[SignalType]:
        result = self.__calls.copy()
        if flush:
            self.flush()
        return result


class SignalWatcher:

    def __init__(self) -> None:
        self.__event = threading.Event()
        self.__signal_desc: typing.Optional[typing.Tuple[SignalSourceProto, Signal]] = None

    def __call__(self, signal_source: SignalSourceProto, signal: Signal, signal_value: typing.Any) -> None:
        self.__event.set()

    def init(self, source: SignalSourceProto, signal: Signal) -> None:
        assert(not self.__signal_desc)
        self.__event.clear()
        source.callback(signal, self)
        self.__signal_desc = (source, signal)

    def reset(self) -> None:
        if self.__signal_desc:
            self.__signal_desc[0].remove_callback(self.__signal_desc[1], self)
            self.__signal_desc = None

    def wait(self, timeout: typing.Union[int, float]) -> None:
        try:
            if not self.__event.wait(timeout):
                raise TimeoutError('Unable to receive a signal in time')
        finally:
            self.reset()


@pytest.fixture
def callbacks_registry(request: pytest.FixtureRequest) -> CallbackRegistry:
    return CallbackRegistry()


@pytest.fixture
def signals_registry(request: pytest.FixtureRequest) -> SignalsRegistry:
    return SignalsRegistry()


@pytest.fixture
def signal_watcher() -> SignalWatcher:
    return SignalWatcher()



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/cron_source.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import calendar
import datetime
import heapq
import typing

from pyknic.lib.verify import verify_value
from pyknic.lib.signals.proto import Signal

from pyknic.lib.tasks.proto import ScheduleSourceProto, ScheduleRecordProto


def cron_token_validator(value_min: int, value_max: int) -> typing.Callable[[int], bool]:
    """ Return function that checks integer value -- that this value resides between lower and upper bounds

    :param value_min: minimum acceptable value
    :param value_max: maximum acceptable value
    """
    def validator(value: int) -> bool:
        return value is None or (value_min <= value <= value_max)
    return validator


class CronSchedule:
    """ This class defines cron-a-like schedule
    """

    __calendar__ = {1: 31, 2: 29, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}

    @verify_value(minute=cron_token_validator(0, 59))
    @verify_value(hour=cron_token_validator(0, 23))
    @verify_value(day_of_month=cron_token_validator(1, 31))
    @verify_value(day_of_week=cron_token_validator(1, 7))
    @verify_value(month=cron_token_validator(1, 12))
    def __init__(
        self,
        minute: typing.Optional[int] = None,
        hour: typing.Optional[int] = None,
        day_of_month: typing.Optional[int] = None,
        day_of_week: typing.Optional[int] = None,
        month: typing.Optional[int] = None
    ):
        """ Create a new schedule

        :param minute: a minute of a schedule (None means every minute)
        :param hour: a hour of a schedule (None means every hour)
        :param day_of_month: a day of a month of a schedule (None means every day)
        :param day_of_week: a day of a week of a schedule (None means every day of week)
        :param month: a month of a schedule (None means every month)
        """

        if month is not None and day_of_month is not None and day_of_month > CronSchedule.__calendar__[month]:
            raise ValueError('Invalid day and month combination')

        self.__minute = minute
        self.__hour = hour
        self.__day_of_month = day_of_month
        self.__day_of_week = day_of_week
        self.__month = month

    def minute(self) -> typing.Optional[int]:
        """ Return scheduler's minute setting
        """
        return self.__minute

    def hour(self) -> typing.Optional[int]:
        """ Return scheduler's hour setting
        """
        return self.__hour

    def day_of_month(self) -> typing.Optional[int]:
        """ Return scheduler's day of month setting
        """
        return self.__day_of_month

    def day_of_week(self) -> typing.Optional[int]:
        """ Return scheduler's day of week setting
        """
        return self.__day_of_week

    def month(self) -> typing.Optional[int]:
        """ Return scheduler's month setting
        """
        return self.__month

    @verify_value(start_datetime=lambda x: x.tzinfo is not None)
    def iterate(self, start_datetime: datetime.datetime) -> typing.Generator[datetime.datetime, None, None]:
        """ Return generator that yields datetime at which something should happen
        """

        def is_start_date(y: int, m: int, d: int) -> bool:
            return y == start_datetime.year and m == start_datetime.month and d == start_datetime.day

        for year, month in self.__month_iterator(start_datetime.year, start_datetime.month):

            current_day = 1
            if year == start_datetime.year and month == start_datetime.month:
                current_day = start_datetime.day  # reset day since a new year/month iteration

            for day in self.__day_iterator(year, month, current_day):

                current_hour = 0
                if is_start_date(year, month, day):
                    current_hour = start_datetime.hour

                for hour in self.__time_iterator(current_hour, 23, self.hour()):
                    current_minute = 0
                    if is_start_date(year, month, day) and hour == start_datetime.hour:
                        current_minute = start_datetime.minute
                    for minute in self.__time_iterator(current_minute, 59, self.minute()):
                        yield datetime.datetime(year, month, day, hour, minute, tzinfo=start_datetime.tzinfo)

    def __month_iterator(
        self, start_year: int, start_month: int
    ) -> typing.Generator[typing.Tuple[int, int], None, None]:
        """ Internal generator that iterates over months
        """
        iter_month = self.month()

        if iter_month is not None:
            if start_month > iter_month:
                start_year += 1
            start_month = iter_month

        next_year = start_year
        next_month = start_month

        while next_year <= datetime.MAXYEAR:
            yield next_year, next_month

            if iter_month is not None:
                next_year += 1
            elif next_month < 12:
                next_month += 1
            else:
                next_year += 1
                next_month = 1

    def __day_iterator(
        self, start_year: int, start_month: int, start_day: int
    ) -> typing.Generator[int, None, None]:
        """ Internal generator that iterates over days
        """

        def weekday(d: int) -> int:
            return calendar.weekday(start_year, start_month, d) + 1  # weekday returns values between 0 and 6

        def switch_day(d: int, max_days: int, dow: typing.Optional[int]) -> typing.Optional[int]:
            if dow is None:
                if d < max_days:
                    return d + 1
                return None

            result = d + 7 + dow - weekday(d)
            if result <= max_days:
                return result
            return None

        days_in_month = calendar.monthrange(start_year, start_month)[1]
        iter_day_of_month = self.day_of_month()
        iter_day_of_week = self.day_of_week()

        if iter_day_of_month is not None:
            if start_day <= iter_day_of_month <= days_in_month:
                if iter_day_of_week is None or weekday(iter_day_of_month) == iter_day_of_week:
                    yield iter_day_of_month
            return

        next_day = start_day
        if iter_day_of_week and weekday(next_day) != iter_day_of_week:
            next_day = switch_day(next_day, days_in_month, iter_day_of_week)  # type: ignore[assignment] # mypy issues

        while next_day is not None:
            yield next_day
            next_day = switch_day(next_day, days_in_month, iter_day_of_week)  # type: ignore[assignment] # mypy issues

    def __time_iterator(
        self, start_time: int, max_time: int, cron_spec: typing.Optional[int]
    ) -> typing.Generator[int, None, None]:
        """ Internal generator that iterates over time
        """
        if cron_spec is not None:
            if start_time <= cron_spec:
                yield cron_spec
            return

        next_time = start_time
        while next_time <= max_time:
            yield next_time
            next_time += 1

    @classmethod
    def from_string(cls, schedule: str) -> 'CronSchedule':
        """ Parse a string and return a schedule
        """
        # TODO: replace with regexp
        return cls.parse_tokens(*tuple(filter(lambda x: len(x) > 0, schedule.strip().split(' '))))

    @classmethod
    def parse_tokens(cls, *str_tokens: str) -> 'CronSchedule':
        """ Parse a string tokens and return a schedule
        """
        if len(str_tokens) != 5:
            raise ValueError('Malformed cron-schedule')

        int_tokens = [int(x) if x != '*' else None for x in str_tokens]

        return cls(
            minute=int_tokens[0],
            hour=int_tokens[1],
            day_of_month=int_tokens[2],
            day_of_week=int_tokens[3],
            month=int_tokens[4]
        )


class CronRecordInitError(Exception):
    """ Raises when the :class:`.CronScheduleRecord` class has initialization issues (object is not initialized or
    initialized twice)
    """
    pass


class CronScheduleRecord:
    """ Represent a single record for a scheduler that has its cron-schedule
    """

    def __init__(self, record: ScheduleRecordProto, schedule: CronSchedule):
        """ Create a new record

        :param record: record to execute
        :param schedule: schedule at which record should be executed
        """
        self.__record = record
        self.__schedule = schedule
        self.__cron_iterator: typing.Optional[typing.Generator[datetime.datetime, None, None]] = None
        self.__next_datetime: typing.Optional[datetime.datetime] = None

    def __init_test(self, is_none: bool) -> None:
        """ Check initialization status

        :param is_none: True if this object must be initialized or False otherwise
        """
        if not is_none and self.__cron_iterator is None:
            raise CronRecordInitError('This record must be initialized before use')
        elif is_none and self.__cron_iterator is not None:
            raise CronRecordInitError('This record is initialized already')

    def init(self, start_datetime: datetime.datetime) -> None:
        """ Initialize this record

        :param start_datetime: this is a start point at which next execution time will be calculated
        """
        self.__init_test(True)
        self.__cron_iterator = self.__schedule.iterate(start_datetime)
        self.__next_datetime = next(self.__cron_iterator)

    def record(self) -> ScheduleRecordProto:
        """ Return a record for execution
        """
        return self.__record

    def next(self) -> datetime.datetime:
        """ Return next execution datetime. May be called several times
        """
        self.__init_test(False)
        return self.__next_datetime  # type: ignore[return-value]  # mypy issue, we have checked this type already

    def commit(self) -> datetime.datetime:
        """ Switch next execution datetime. After this method call the :meth:`.CronScheduleRecord.next` method
        will change value
        """
        self.__init_test(False)
        self.__next_datetime = next(self.__cron_iterator)  # type: ignore[arg-type]  # mypy issue
        return self.__next_datetime

    def schedule(self, start_datetime: datetime.datetime) -> typing.Generator[datetime.datetime, None, None]:
        """ Return a new generator of an internal schedule
        """
        return self.__schedule.iterate(start_datetime)

    def __lt__(self, other: 'CronScheduleRecord') -> bool:
        """ Compare two :class:`.CronScheduleRecord` objects. Is used for ordering objects
        """
        return self.next() < other.next()


class CronTaskSource(ScheduleSourceProto):
    """ The :class:`.ScheduleSourceProto` implementation that holds and triggers cron-a-like records

    :note: This class is NOT thread safe!
    """

    polling_update = Signal(int)  # this signal is related to the :meth:`.CronTaskSource.poll` method, it will be
    # emitted when a previously returned value is not valid anymore. It may happen when a new record appear,
    # or an upcoming record is removed

    def __init__(self) -> None:
        """ Create a new task source
        """
        ScheduleSourceProto.__init__(self)

        self.__records: typing.List[CronScheduleRecord] = []
        heapq.heapify(self.__records)

    def __now(self) -> datetime.datetime:
        """ Return datetime object that will be treated as "now" (as it is in UTC timezone)
        """
        return datetime.datetime.now(datetime.timezone.utc)

    def submit_record(self, record: CronScheduleRecord) -> None:
        """ Append a record to this source
        """

        # if new record updates last poll timeout a signal polling_update will be sent
        now = self.__now()
        record.init(now)

        current_record_dt = record.next()
        previous_record_dt = self.__records[0].next() if self.__records else None

        heapq.heappush(self.__records, record)

        if previous_record_dt is None or previous_record_dt > current_record_dt:
            delta = current_record_dt - now
            self.emit(self.polling_update, delta.seconds)

    def __emit_records(self) -> None:
        """ Check whether there are records that are ready to be executed (their time has come)
        """
        records_expired = True
        now = self.__now()
        records_to_emit = []

        while records_expired and self.__records:
            cron_record = self.__records[0]
            if cron_record.next() <= now:
                records_to_emit.append(heapq.heappop(self.__records))
            else:
                records_expired = False

        for cron_record in records_to_emit:
            cron_record.commit()
            self.emit(self.task_scheduled, cron_record.record())
            heapq.heappush(self.__records, cron_record)

    def poll(self) -> typing.Optional[int]:
        """ Trigger emitting records which time has come. And return number of seconds to wait for next records

        :return: number of seconds to wait if there is at least one record to wait, and return None otherwise
        """
        " ScheduleSourceProto.task_scheduled will be emitted if require "

        if not self.__records:
            return None

        self.__emit_records()
        cron_record = self.__records[0]
        delta = self.__now() - cron_record.next()
        return delta.seconds

    def records(self) -> typing.Generator[CronScheduleRecord, None, None]:
        """ Return generator that yields record this source has
        """
        for i in self.__records.copy():
            yield i

    def discard_record(self, record: CronScheduleRecord) -> None:
        """ Remove a record from this source

        :param record: a record to remove from this source
        """
        for i in range(len(self.__records)):
            if self.__records[i] == record:
                self.__records.pop(i)
                if self.__records:
                    delta = self.__now() - self.__records[0].next()
                    self.emit(self.polling_update, delta.seconds)
                return
        raise ValueError('Unable to find a record')



# -*- coding: utf-8 -*-
# pyknic/tasks/fastapi/server.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code
# TODO: tweak uvicorn logger settings

import asyncio
import uuid

import fastapi
import typing
import uvicorn

from dataclasses import dataclass

from pyknic.lib.datalog.proto import DatalogProto
from pyknic.lib.tasks.scheduler.chain_source import ChainedTask, __default_chained_tasks_registry__
from pyknic.lib.fastapi.apps_registry import __default_fastapi_apps_registry__
from pyknic.lib.registry import register_api

from pyknic.lib.log import Logger


@dataclass
class FastAPIServer:
    fastapi_app: fastapi.FastAPI
    uvicorn_server: uvicorn.Server


@register_api(__default_chained_tasks_registry__, "fastapi-init")
class FastAPIInitTask(ChainedTask):

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        Logger.info('Check dependencies')

        self.wait_for('log_task')
        config = self.wait_for('config_task').result  # type: ignore[union-attr]

        Logger.info('Prepare fastAPI')

        config_section = config["pyknic"]["fastapi"]

        docs_url = '/docs' if bool(config_section["swagger"]) else None
        redoc_url = '/redoc' if bool(config_section["redoc"]) else None
        openapi_url = '/openapi.json' if bool(config_section["openapi"]) else None

        app = fastapi.FastAPI(docs_url=docs_url, redoc_url=redoc_url, openapi_url=openapi_url)  # TODO: check options!
        uvicorn_config = uvicorn.Config(  # TODO: check options!
            app, host=str(config_section["uvicorn_host"]), port=int(config_section["uvicorn_port"])
        )
        uvicorn_server = uvicorn.Server(uvicorn_config)

        self.save_result(FastAPIServer(fastapi_app=app, uvicorn_server=uvicorn_server))

        Logger.info('FastAPI prepared')

    def task_name(self) -> typing.Optional[str]:
        """ The :meth:`.ChainedTask.task_name` method implementation
        """
        return 'fastAPI-init'

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ The :meth:`.ChainedTask.dependencies` method implementation
        """
        return {"log_task", "config_task"}


@register_api(__default_chained_tasks_registry__, "fastapi-loader")
class FastAPILoaderTask(ChainedTask):

    def __init__(self, datalog: DatalogProto, api_id: str, uid: uuid.UUID):
        ChainedTask.__init__(self, datalog, api_id, uid)

        # just to be sure that important info will not be compacted by the mighty GC
        self.__loaded_apps: typing.List[typing.Any] = []

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        Logger.info('Populate fastAPI server')

        fastapi_init = self.wait_for('fastapi-init')
        config_result = self.wait_for('config_task')
        gettext_result = self.wait_for('gettext_task')

        assert(fastapi_init)
        assert(config_result)
        assert(gettext_result)

        config = config_result.result

        apps_enabled = [str(x) for x in config["pyknic"]["fastapi"]["apps"]]
        apps_enabled.sort()

        for app_id in apps_enabled:
            Logger.info(f'Reading the app "{app_id}"')
            fastapi_app_cls = __default_fastapi_apps_registry__.get(app_id)
            created_app = fastapi_app_cls.create_app(fastapi_init.result.fastapi_app, config, gettext_result.result)
            if created_app is not None:
                self.__loaded_apps.append(created_app)

    def task_name(self) -> typing.Optional[str]:
        """ The :meth:`.ChainedTask.task_name` method implementation
        """
        return 'fastAPI-loader'

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ The :meth:`.ChainedTask.dependencies` method implementation
        """
        return {"fastapi-init", "gettext_task"}


@register_api(__default_chained_tasks_registry__, "fastapi-server")
class FastAPIServerTask(ChainedTask):

    def __init__(self, datalog: DatalogProto, api_id: str, uid: uuid.UUID):
        ChainedTask.__init__(self, datalog=datalog, api_id=api_id, uid=uid)

        self.__loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        self.__uvicorn_server: typing.Optional[uvicorn.Server] = None

    async def __stop_request(self) -> None:
        assert(self.__uvicorn_server)

        self.__uvicorn_server.should_exit = True
        await self.__uvicorn_server.shutdown()

    def start(self) -> None:
        """ The :meth:`.TaskProto.start` method implementation
        """
        Logger.info('Starting fastAPI')
        assert(self.__loop)

        fastapi_init = self.wait_for('fastapi-init')
        assert(fastapi_init)

        self.__uvicorn_server = fastapi_init.result.uvicorn_server
        self.__loop.run_until_complete(self.__uvicorn_server.serve())

        Logger.info('FastAPI stopped')

    def stop(self) -> None:
        Logger.info('Generate fastAPI stop request')
        future = asyncio.run_coroutine_threadsafe(self.__stop_request(), self.__loop)
        future.result()
        Logger.info('FastAPI stop requested')

    def task_name(self) -> typing.Optional[str]:
        """ The :meth:`.ChainedTask.task_name` method implementation
        """
        return 'fastAPI-server'

    @classmethod
    def dependencies(cls) -> typing.Optional[typing.Set[str]]:
        """ The :meth:`.ChainedTask.dependencies` method implementation
        """
        return {"fastapi-loader"}



# -*- coding: utf-8 -*-
# pyknic/lib/crypto/proto.py
#
# Copyright (C) 2018-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from abc import ABCMeta, abstractmethod

from cryptography.hazmat.primitives.hashes import HashAlgorithm as C10yHashAlgorithm

from pyknic.lib.io import IOGenerator, IOProducer
from pyknic.lib.verify import verify_value


class HasherProto(metaclass=ABCMeta):
    """This is a base class for all generalized hashers."""

    @abstractmethod
    def update(self, source: IOProducer) -> IOGenerator:
        """Update digest and yield all the received data.

        :param source: a data to update digest
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def digest(self) -> bytes:
        """Return generated digest."""
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def c10y_algorithm(self) -> typing.Optional[C10yHashAlgorithm]:
        """If available -- return object of the "cryptography" module which implements the hash algorithm
        """
        return None


class CipherProto(metaclass=ABCMeta):
    """ This class is a generalization of ciphers (now it is AES cipher only)
    """

    @classmethod
    @abstractmethod
    def algo_block_size(cls) -> typing.Optional[int]:
        """Return cipher's block size

        :return: int (in bytes) or None if cipher is a stream cipher (not a block cipher)
        """
        raise NotImplementedError('This method is abstract')

    @classmethod
    @abstractmethod
    def key_size(cls) -> typing.Optional[int]:
        """Return a size of a key this cipher require

        :return: int (in bytes) or None if cipher is able to use a key with any length
        """
        raise NotImplementedError('This method is abstract')

    @classmethod
    @abstractmethod
    def create_encryptor(cls, key: bytes) -> 'CipherProto':
        """Create a new cipher object that is able to encrypt data (decryption may not be possible and
        calling to the :meth:`.CipherProto.decrypt` should be avoided)

        :param key: key to use for encryption
        """

        raise NotImplementedError('This method is abstract')

    @classmethod
    @abstractmethod
    def create_decryptor(cls, key: bytes, decryptor_init_data: typing.Any) -> 'CipherProto':
        """Create a new cipher object that is able to decrypt data (encryption may not be possible and
        calling to the :meth:`.CipherProto.encrypt` should be avoided)

        :param key: key to use for decryption
        :param decryptor_init_data: initialization data with which cipher was created.
        See the :meth:`.CipherProto.decryptor_init_data` method
        """

        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def decryptor_init_data(self) -> typing.Any:
        """Initialization data with which this cipher was created. This data may contain some random bytes (like nonce
        or initialization vector) that are crucial for decryption. This data is safe to transmit along with secured
        message.

        :note: result of this method should be able to serialize to JSON
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def encrypt(self, data: IOProducer) -> IOGenerator:
        """Encrypt the given data

        :param data: data to encrypt
        """
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def decrypt(self, data: IOProducer) -> IOGenerator:
        """Decrypt the given data

        :param data: data to decrypt
        """
        raise NotImplementedError('This method is abstract')


class BlockPaddingProto(metaclass=ABCMeta):
    """ Padding/reverse padding class prototype
    """

    @abstractmethod
    @verify_value(block_size=lambda x: x > 0)
    def pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ Pad given data to given size

        :param data: data to pad
        :param block_size: size to pad
        :return: bytes
        """
        raise NotImplementedError("This method is abstract")

    @abstractmethod
    @verify_value(block_size=lambda x: x > 0)
    def undo_pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """Remove pads and return original data

        :param data: data to remove pads from
        :param block_size: size data aligned to
        :return: bytes
        """
        raise NotImplementedError("This method is abstract")



# -*- coding: utf-8 -*-
# pyknic/lib/bellboy/console.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

import rich.console
import rich.padding
import rich.prompt
import rich.table
import rich.text

from pyknic.lib.bellboy.app import BellboyCLIError
from pyknic.lib.fastapi.models.lobby import LobbyStrFeedbackResult, LobbyKeyValueFeedbackResult, LobbyCommandResult
from pyknic.lib.fastapi.models.lobby import LobbyListValueFeedbackResult
from pyknic.lib.fastapi.models.base import NullableModel


class BellboyConsole:
    """ Wraps and unify style and CLI behavior"""

    def __init__(self) -> None:
        """ Create a new console renderer
        """

        self.__console = rich.console.Console()
        self.__last_prompt: typing.Optional[str] = None

    def ask(self, question: str | None = None, password: bool = False) -> str:
        """ Ask a user for an input

        :param question: if defined then this question will be printed before prompt
        :param password: whether prompt should echo an input or not
        """
        if question:
            self.__console.print(rich.text.Text(question))
        result = rich.prompt.Prompt.ask(password=password)
        self.__last_prompt = result.strip()
        return self.__last_prompt

    def critical(self, raised_exception: BellboyCLIError) -> None:
        """Process raised exception -- print error message and reraise the one.

        :param raised_exception: raised exception
        """
        self.__console.print(
            rich.text.Text(f'Fatal error spotted: "{str(raised_exception)}". Quiting...', style='red bold')
        )
        raise raised_exception

    def error(self, message: str) -> None:
        """Show error message

        :param message: error message
        """
        self.__console.print(rich.text.Text(f'Error! {message}', style='red'))

    def str_feedback(self, feedback: LobbyStrFeedbackResult) -> None:
        """Process simple result.

        :param feedback: sting-based result
        """
        self.__console.print(rich.text.Text(feedback.str_result, style='bold'))

    def kv_feedback(self, feedback: LobbyKeyValueFeedbackResult) -> None:
        """Process key-value result.

        :param feedback: dictionary result
        """
        table = rich.table.Table(title='Values received')
        table.add_column('Name', style='bold')
        table.add_column('Values', style='green', justify='right')

        table_keys = list(feedback.kv_result.keys())
        table_keys.sort()

        for k in table_keys:
            table.add_row(k, str(feedback.kv_result[k]))

        self.__console.print(table)

    def null_feedback(self, feedback: NullableModel) -> None:
        """Process null result.

        :param feedback: None as is =)
        """
        self.__console.print(rich.text.Text('<command succeeded without feedback>'))

    def list_feedback(self, feedback: LobbyListValueFeedbackResult) -> None:
        """Process null result.

        :param feedback: None as is =)
        """
        self.__console.print(rich.text.Text('Values received:'))
        for i in feedback.list_result:
            self.__console.print(rich.padding.Padding.indent(i, 4))

    def process_result(
        self,
        command_result: LobbyCommandResult
    ) -> None:
        """Print a result in a stylish way

        :param command_result: command result to print
        """
        if isinstance(command_result, LobbyStrFeedbackResult):
            self.str_feedback(command_result)
        elif isinstance(command_result, LobbyKeyValueFeedbackResult):
            self.kv_feedback(command_result)
        elif isinstance(command_result, LobbyListValueFeedbackResult):
            self.list_feedback(command_result)
        elif isinstance(command_result, NullableModel):
            self.null_feedback(command_result)
        else:
            raise BellboyCLIError('Unknown command result spotted! Check logs')



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/lobby_fingerprint.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import base64
import secrets

from cryptography.hazmat.primitives import hashes, hmac

from pyknic.lib.verify import verify_value


class LobbyFingerprint:
    """This class wraps fingerprint routine and HMAC-signing things."""

    @verify_value(fingerprint=lambda x: len(x) == LobbyFingerprint.fingerprint_bytes())
    def __init__(self, fingerprint: bytes) -> None:
        """Create an object that may implement some functions with fingerprint.

        :param fingerprint: fingerprint to use
        """
        self.__fingerprint = fingerprint

    @classmethod
    def from_model(
        cls, fingerprint_model: 'LobbyFingerprintModel'  # type: ignore[name-defined]  # noqa: F821  # recurse issue
    ) -> 'LobbyFingerprint':
        """ Convert model to this class

        :param fingerprint_model: model to convert
        """
        return LobbyFingerprint.deserialize(
                fingerprint_model.fingerprint.encode('ascii')
        )

    @classmethod
    def fingerprint_bytes(cls) -> int:
        """Return fixed number of bytes that this fingerprint has. In HMAC-SHA256 it must be the same size
        as 32 bytes (256 bit)
        """
        return 32

    def __str__(self) -> str:
        """Return fingerprint as a base64 encoded string."""
        return base64.b64encode(self.__fingerprint).decode('ascii')

    def sign(self, data_to_sign: bytes, encode_base64: bool = False) -> bytes:
        """Return signature of data.

        :param data_to_sign: data to sign
        :param encode_base64: whether signature should be return as is or as base64 encoded bytes
        """
        h = hmac.HMAC(self.__fingerprint, hashes.SHA256())
        h.update(data_to_sign)
        result = h.finalize()
        if encode_base64:
            result = base64.b64encode(result)
        return result

    @classmethod
    def generate_fingerprint(cls) -> 'LobbyFingerprint':
        """Return randomly generated sequence of bytes."""
        # TODO: check if it is secure enough (at the first glance it is ok but still)
        fingerprint = secrets.token_bytes(cls.fingerprint_bytes())
        return LobbyFingerprint(fingerprint)

    @classmethod
    def serialized_length(cls) -> int:
        """Return number of bytes required to represent base64 encoded fingerprint. In base64 every 3 bytes are
        encoded with 4 symbols so the target size is 44
        """
        return 44

    @classmethod
    def deserialize(cls, fingerprint: bytes) -> 'LobbyFingerprint':
        """Return fingerprint object from base64 encoded string.

        :param fingerprint: base64 encoded fingerprint
        """
        return LobbyFingerprint(base64.b64decode(fingerprint))



# -*- coding: utf-8 -*-
# pyknic/tasks/fastapi/word_games/cities.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

import fastapi
import pathlib
import random
import typing

from pyknic.lib.config import Config
from pyknic.lib.fastapi.models.tg_bot_types import Update
from pyknic.lib.fastapi.base import TgBotBaseFastAPIApp, TgBotResponseType
from pyknic.lib.gettext import GetTextWrapper
from pyknic.path import root_path


class TGBotCityGame(TgBotBaseFastAPIApp):

    __default_city_file__ = root_path / 'data' / 'cities_ru.txt'
    # Original file is made with this resource -- "https://ru.wikinews.org/wiki/Категория:Города_по_алфавиту"

    @classmethod
    def create_app(cls, fastapi_app: fastapi.FastAPI, config: Config, translations: GetTextWrapper) -> typing.Any:
        raise TypeError('This class should be created directly')

    @classmethod
    def bot_path(cls, config: Config) -> str:
        raise TypeError('This class should be created directly')

    async def tgbot_webhook_request(self, request: fastapi.Request) -> TgBotResponseType:
        raise TypeError('This class should be created directly')

    def __init__(self, config: Config, translations: GetTextWrapper):
        TgBotBaseFastAPIApp.__init__(self, config, translations)

        self.__cities: typing.Set[str] = set()
        self.__sessions: typing.Dict[int, typing.List[str]] = dict()

        self.__update_cities(self.__default_city_file__)

    def __update_cities(self, file_path: str | pathlib.Path) -> None:
        with open(file_path) as f:
            self.__cities.update((x.strip() for x in f.readlines() if len(x.strip())))

    def __allowed_city(self, city_name: str, user_id: int) -> typing.Optional[str]:
        mentioned_cities = self.__sessions[user_id]

        limited_cities = [
            x for x in self.__cities if x.startswith(city_name[-1]) and x != city_name and x not in mentioned_cities
        ]

        if len(limited_cities) > 0:
            return random.choice(limited_cities)
        elif len(limited_cities) == 0 and len(city_name) > 1:
            return self.__allowed_city(city_name[:-1], user_id)
        return None

    async def reset(self, id_: int, _: Update) -> None:
        self.__sessions.setdefault(id_, []).clear()

    def last_city(self, id_: int) -> str | None:
        cities = self.__sessions.setdefault(id_, [])
        if cities:
            return cities[-1]
        return None

    async def process_command(self, command: str, tg_update: Update) -> TgBotResponseType | None:
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)

        last_city_name = self.last_city(tg_update.message.from_.id_)
        lang = self.user_lang(tg_update.message.from_)

        if command == "/help":
            if last_city_name is None:
                return self.reply(tg_update.message, lang.gettext("Name any city you know! We have just started!"))

            city_name = self.__allowed_city(last_city_name, tg_update.message.from_.id_)
            if city_name is not None:
                return self.reply(
                    tg_update.message,
                    lang.gettext('The "%(cn)s" city is an option') % {'cn': city_name.capitalize()}
                )
            else:
                return self.reply(tg_update.message, lang.gettext('Know nothing more. One more game?'))
        elif command == "/stats":
            city_count = len(self.__sessions[tg_update.message.from_.id_])
            return self.reply(
                tg_update.message,
                lang.ngettext(
                    'We have mentioned %(cc)i city',
                    'We have mentioned %(cc)i cities',
                    city_count
                ) % {'cc': city_count}
            )

        return self.reply(tg_update.message, lang.gettext('Unknown command -- "%(c)s"') % {'c': command})

    async def process_message(self, tg_update: Update) -> TgBotResponseType:
        assert(tg_update.message is not None)
        assert(tg_update.message.from_ is not None)
        assert(tg_update.message.text is not None)

        mentioned_cities = self.__sessions.setdefault(tg_update.message.from_.id_, [])
        msg = tg_update.message.text.lower().strip()

        lang = self.user_lang(tg_update.message.from_)

        if len(msg) == 0:
            return self.reply(tg_update.message, lang.gettext('Please repeat'))

        if msg not in self.__cities:
            return self.reply(
                tg_update.message,
                lang.gettext('Unknown city -- "%(cn)s"') % {'cn': msg.capitalize()}
            )

        if msg in mentioned_cities:
            return self.reply(
                tg_update.message,
                lang.gettext('The "%(cn)s" city has been mentioned already') % {'cn': msg.capitalize()}
            )

        if len(mentioned_cities) > 0:
            prev_city = mentioned_cities[-1]
            check_city = self.__allowed_city(prev_city, tg_update.message.from_.id_)

            if check_city is None:
                return self.reply(tg_update.message, lang.gettext('No more cities left. I -- loose; you -- win!'))

            if check_city[0] != msg[0]:
                return self.reply(
                    tg_update.message,
                    lang.gettext("The \"%(cn)s\" city doesn't starts with the %(l)s") % {
                        'cn': msg.capitalize(), 'l': check_city[0]
                    }
                )

        mentioned_cities.append(msg)
        answer = self.__allowed_city(msg, tg_update.message.from_.id_)

        if answer is None:
            return self.reply(tg_update.message, lang.gettext('No more cities left. I -- loose; you -- win!'))

        mentioned_cities.append(answer)
        return self.reply(tg_update.message, text=answer.capitalize())



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/apps_registry.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code

import fastapi
import typing

from abc import ABCMeta, abstractmethod

from pyknic.lib.config import Config
from pyknic.lib.registry import APIRegistry
from pyknic.lib.gettext import GetTextWrapper


__default_fastapi_apps_registry__ = APIRegistry()


class FastAPIAppProto(metaclass=ABCMeta):
    """ Default prototype for loadable by the FastAPI task
    """

    @classmethod
    @abstractmethod
    def create_app(cls, fastapi_app: fastapi.FastAPI, config: Config, translations: GetTextWrapper) -> typing.Any:
        """ Insert endpoints to a fastapi_app that about to be started

        :param fastapi_app: application to update
        :param config: pyknic's global configuration
        :param translations: preloaded localization

        :return: anything that should be saved during the work
        """
        raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/__init__.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

from pyknic.lib.integrated_commands.backup import BellBoyBackupCommand, BellBoyArchiveValidateCommand
from pyknic.lib.integrated_commands.backup import BellBoyRestoreCommand
from pyknic.lib.integrated_commands.list_logins import ListLoginsCommand
from pyknic.lib.integrated_commands.login import LoginCommand
from pyknic.lib.integrated_commands.logout import LogoutCommand
from pyknic.lib.integrated_commands.logout_all import LogoutAllCommand
from pyknic.lib.integrated_commands.ping import LobbyPingCommand, BellBoyPingCommand
from pyknic.lib.integrated_commands.resources import LobbyResourcesCommand, BellBoyResourcesCommand

__all__ = [
    'BellBoyBackupCommand',
    'BellBoyArchiveValidateCommand',
    'BellBoyRestoreCommand',
    'ListLoginsCommand',
    'LoginCommand',
    'LogoutCommand',
    'LogoutAllCommand',
    'LobbyPingCommand',
    'BellBoyPingCommand',
    'LobbyResourcesCommand',
    'BellBoyResourcesCommand',
]



# -*- coding: utf-8 -*-
# pyknic/lib/uri.py
#
# Copyright (C) 2017-2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing
from urllib.parse import urlsplit, parse_qs, urlencode

from pyknic.lib.property import TypedDescriptor


URIQueryParameterType = typing.TypeVar('URIQueryParameterType')


class URIQueryInvalidSingleParameter(Exception):
    pass


class URI:
    """Class that represent URI as it is described in RFC 3986.

    https://datatracker.ietf.org/doc/html/rfc3986
    """

    scheme = TypedDescriptor(str)
    username = TypedDescriptor(str)
    password = TypedDescriptor(str)
    hostname = TypedDescriptor(str)
    port = TypedDescriptor(int, value_verifier=lambda x: x is None or (0 <= x <= 65535))
    path = TypedDescriptor(str)
    query = TypedDescriptor(str)
    fragment = TypedDescriptor(str)

    def __init__(
        self,
        *,
        scheme: typing.Optional[str] = None,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        hostname: typing.Optional[str] = None,
        port: typing.Optional[int] = None,
        path: typing.Optional[str] = None,
        query: typing.Optional[str] = None,
        fragment: typing.Optional[str] = None
    ):
        """Create a new URI."""
        self.scheme = scheme
        self.username = username
        self.password = password
        self.hostname = hostname
        self.port = port
        self.path = path.lstrip('/') if path else None
        self.query = query
        self.fragment = fragment

    def __str__(self) -> str:
        """ Return string that represents this URI."""
        # note the urlunsplit function in unpredictable because of this:
        #   >>> urlunsplit(('foo', '', '/path', None, None))
        #   'foo:/path'
        #   >>> urlunsplit(('http', '', '/path', None, None))
        #   'http:///path'

        def default_fn(value: typing.Optional[str], default_value: str) -> str:
            return value if value is not None else default_value

        def prefixed_value(value: typing.Optional[str], prefix: str, default_value: str) -> str:
            return prefix + value if value else default_value

        auth = default_fn(self.username, "")
        if auth and self.password:
            auth += ':' + self.password
        if auth:
            auth += '@'

        scheme = default_fn(self.scheme, "")
        if scheme:
            scheme += '://'

        netloc = default_fn(self.hostname, "")
        if netloc and self.port:
            netloc += ':' + str(self.port)

        path = prefixed_value(self.path, "/", "")
        query = prefixed_value(self.query, "?", "")
        fragment = prefixed_value(self.fragment, "#", "")

        return f'{scheme}{auth}{netloc}{path}{query}{fragment}'

    @classmethod
    def parse(cls, uri: str) -> 'URI':
        """Parse a string and return a new URI."""
        uri_components = urlsplit(uri)

        def adapter_str_fn(value: typing.Optional[str]) -> typing.Optional[str]:
            if value is not None and len(value) > 0:
                return value
            return None

        return cls(
            scheme=adapter_str_fn(uri_components.scheme),
            username=adapter_str_fn(uri_components.username),
            password=adapter_str_fn(uri_components.password),
            hostname=adapter_str_fn(uri_components.hostname),
            port=uri_components.port,
            path=adapter_str_fn(uri_components.path),
            query=adapter_str_fn(uri_components.query),
            fragment=adapter_str_fn(uri_components.fragment),
        )


class URIQuery:
    """ Represent a query component of URI. Any parameter may be present more than one time
    """
    # TODO: add quote/ unqoute to replace %XX

    def __init__(self) -> None:
        """ Create new query component
        """
        self.__query: typing.Dict[str, typing.List[str]] = dict()

    def update(self, name: str, value: typing.Union[str, typing.Tuple[str, ...]], append: bool = False) -> None:
        """Add/append/replace parameter in this query.

        :param name: parameter name to change
        :param value: parameter value to set, tuple for multiple values
        :param append: whether to append or replace parameter
        """
        value_list = list(value) if isinstance(value, tuple) else [value]
        if append and name in self.__query:
            value_list = self.__query[name] + value_list

        self.__query[name] = value_list

    def remove(self, name: str) -> None:
        """ Remove the specified parameter from this query.

        :param name: name of a parameter to remove
        """
        if name in self.__query:
            self.__query.pop(name)

    def __contains__(self, item: str) -> bool:
        """ Check if this query has the specified parameter

        :param item: parameter name to check
        """
        return item in self.__query

    def __str__(self) -> str:
        """ Encode parameters from this query, so it can be use in URI
        """
        parameters_keys = list(self.__query.keys())
        parameters_keys.sort()

        return '&'.join((urlencode({x: self.__query[x]}, True)) for x in parameters_keys)

    def __getitem__(self, item: str) -> typing.Tuple[str, ...]:
        """ Return all values for a signle parameter

        :param item: parameter name to retrieve
        """
        return tuple(self.__query[item])

    def __iter__(self) -> typing.Generator[str, None, None]:
        """ Iterate over parameters names
        """
        parameters_keys = list(self.__query.keys())
        parameters_keys.sort()
        yield from parameters_keys

    def parameters(self) -> typing.Generator[typing.Tuple[str, typing.Tuple[str, ...]], None, None]:
        """ Iterate over items, a pair of component name and component value will be raised
        """
        yield from ((x, tuple(y)) for x, y in self.__query.items())

    def single_parameter(
        self, name: str, type_adapter: typing.Callable[[str], URIQueryParameterType]
    ) -> URIQueryParameterType:
        """Cast and return value of a single parameter of this query.

        :param name: parameter name to retrieve
        :param type_adapter: function to convert strings to a type (for example -- int, float and so on)
        """
        if name not in self:
            raise URIQueryInvalidSingleParameter(f'The "{name}" parameter is not defined')

        str_multiple_parameters = self[name]
        if len(str_multiple_parameters) != 1:
            raise URIQueryInvalidSingleParameter(f'The "{name}" parameter is set multiple times')

        str_parameter = str_multiple_parameters[0]
        if str_parameter == '':
            if type_adapter is bool:
                return True  # type: ignore[return-value]
            raise URIQueryInvalidSingleParameter(f'The "{name}" parameter has null value')

        return type_adapter(str_parameter)

    @classmethod
    def parse(cls, query_str: str) -> 'URIQuery':
        """ Parse string that represent query component from URI

        :param query_str: string without '?'-sign
        :type query_str: str
        """
        result = cls()

        if not query_str:
            return result

        parsed_query = parse_qs(query_str, keep_blank_values=True, strict_parsing=True)
        for name, value in parsed_query.items():
            result.update(name, tuple(value))

        return result



# -*- coding: utf-8 -*-
# pyknic/lib/io/stats.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.


import time
import typing

from pyknic.lib.io import IOGenerator, IOProducer


class GeneratorStats:
    """This IOProcessor is used for some statistics calculation.
    """

    def __init__(self) -> None:
        """Create and initialize statistics
        """
        self.__start_time: typing.Optional[float] = None
        self.__finish_time: typing.Optional[float] = None
        self.__bytes = 0

    def rate(self) -> float:
        """Return number of processed bytes from an original source per second. This method is available only after
        the generator has been finished.
        """
        if self.__finish_time is None:
            raise RuntimeError('The generator has not been finished')

        elapsed = self.__finish_time - self.__start_time  # type: ignore[operator]
        return float(self.__bytes) / elapsed

    def bytes(self) -> int:
        """Return number of processed bytes from an original source. This method is available only after the generator
        has been finished.
        """
        if self.__finish_time is None:
            raise RuntimeError('The generator has not been finished')
        return self.__bytes

    def process(self, source: IOProducer) -> IOGenerator:
        """Process a source and calculate statistics.

        :param source: source which statistics should be calculated
        """
        if self.__start_time is None:
            self.__start_time = time.monotonic()

        for i in source:
            self.__bytes += len(i)
            yield i

        self.__finish_time = time.monotonic()



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/virtual_dir.py
#
# Copyright (C) 2017-2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
import typing

from pyknic.lib.registry import register_api
from pyknic.lib.uri import URI, URIQuery
from pyknic.lib.io.clients.virtual_dir import VirtualDirectoryClient
from pyknic.lib.io.clients.collection import __default_io_clients_registry__
from pyknic.lib.verify import verify_value
from pyknic.lib.io.aio_wrapper import IOThrottler, cag


@register_api(__default_io_clients_registry__, "file")
class LocalClient(VirtualDirectoryClient):
    """Local files implementation of :class:`.IOClientProto`."""

    def __init__(self, uri: URI) -> None:
        """Create a new client

        :param uri: URI with which this client should be created. If uri has the "path" attribute, then this path
        will be used as a start point
        """

        VirtualDirectoryClient.__init__(self, uri)

        self.__block_size = 4096

        if uri.query:
            query = URIQuery.parse(uri.query)

            if 'block_size' in query:
                block_size = query.single_parameter('block_size', int)
                if block_size < 4096:
                    raise ValueError(f'Block size must be greater than {4096} bytes, got {block_size}')
                self.__block_size = block_size

        if uri.path is not None:
            self.__change_directory(pathlib.PosixPath(uri.path))

    def current_directory(self) -> str:
        """The :meth:`.IOClientProto.current_directory` method implementation."""
        return str(self.session_path())

    def __change_directory(self, path: pathlib.PosixPath) -> pathlib.PosixPath:
        """Change current session directory to the specified one

        :param path: new session directory
        """
        if not path.is_absolute():
            path = self.session_path() / path

        if not pathlib.PosixPath(path).is_dir():
            raise NotADirectoryError(f'No such directory: {str(path)}')
        return self.session_path(pathlib.PosixPath(path))

    async def change_directory(self, path: str) -> str:
        """The :meth:`.IOClientProto.change_directory` method implementation."""
        return str(self.__change_directory(pathlib.PosixPath(path)))

    async def list_directory(self) -> typing.Tuple[str, ...]:
        """The :meth:`.IOClientProto.list_directory` method implementation."""
        return tuple(x.name for x in self.session_path().iterdir())

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def make_directory(self, directory_name: str) -> None:
        """The :meth:`.IOClientProto.make_directory` method implementation."""
        path = pathlib.PosixPath(self.session_path()) / directory_name
        path.mkdir(exist_ok=False, parents=False)

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def remove_directory(self, directory_name: str) -> None:
        """The :meth:`.IOClientProto.remove_directory` method implementation."""
        path = pathlib.PosixPath(self.session_path()) / directory_name
        path.rmdir()

    @verify_value(from_fo=lambda x: x.seekable())
    @verify_value(to_fo=lambda x: x.seekable())
    async def __copy(self, from_fo: typing.IO[bytes], to_fo: typing.IO[bytes]) -> None:
        """Copy files from one to another.

        :param from_fo: file to copy
        :param to_fo: file to copy to
        """
        from_fo.seek(0)
        to_fo.truncate(0)
        to_fo.seek(0)

        await cag(IOThrottler().async_copier(from_fo, to_fo, block_size=self.__block_size))

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    async def upload_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """The :meth:`.IOClientProto.upload_file` method implementation."""
        path = self.entry_path(remote_file_name)
        with open(str(path), mode='wb') as f_remote:
            await self.__copy(local_file_obj, f_remote)

    @verify_value(file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def remove_file(self, file_name: str) -> None:
        """The :meth:`.IOClientProto.remove_file` method implementation."""
        path = self.entry_path(file_name)
        path.unlink()

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    async def receive_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """The :meth:`.IOClientProto.receive_file` method implementation."""
        path = str(self.entry_path(remote_file_name))
        with open(path, mode='rb') as f_remote:
            await self.__copy(f_remote, local_file_obj)

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def file_size(self, remote_file_name: str) -> int:
        """The :meth:`.IOClientProto.file_size` method implementation."""
        return self.entry_path(remote_file_name).stat().st_size



# -*- coding: utf-8 -*-
# pyknic/lib/fastapi/lobby_commands/backup.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: write tests for the code
# TODO: make the remote LobbyCommandHandler command for backup, restore and validate

import base64
import shlex
import subprocess
import typing

import pydantic
import pydantic_settings

from pyknic.lib.io import IOGenerator
from pyknic.lib.io.aio_wrapper import IOThrottler, cag
from pyknic.lib.backup.archive_v1 import HashMethod, CompressionMode, BackupArchiveV1
from pyknic.lib.bellboy.app import BellBoyCommandHandler, register_bellboy_command
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyKeyValueFeedbackResult, LobbyStrFeedbackResult


class BackupSource(pydantic_settings.CliMutuallyExclusiveGroup):
    """ These settings define which data should be backed up
    """

    command: typing.Optional[str] = pydantic.Field(
        default=None,
        description='a command to execute, which stdout will be backed up'
    )
    files: typing.Optional[typing.List[str]] = pydantic.Field(
        default=None,
        description='a static files/directories list to backup'
    )
    files_command: typing.Optional[str] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('files-command'),
        default=None,
        description='a command to execute and every line of the stdout will be treated as a file to backup'
    )


class BackupCommandModel(pydantic.BaseModel):
    """ These settings define main backup options
    """

    hash_algorithms: typing.Optional[typing.List[HashMethod]] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('hash-algorithms'),
        default=None,
        description='list of hash algorithms to use'
    )

    throttling: typing.Optional[int] = pydantic.Field(
        default=None,
        description='target data-rate (bytes per second) for saving archive'
    )

    compression: typing.Optional[CompressionMode] = pydantic.Field(
        default=None,
        description='type of compression to use'
    )

    cipher_name: typing.Optional[str] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('cipher-name'),
        default=None,
        description='cipher to use (will be enabled only if the encryption-key was set), '
        'like: AES-256-CBC (which is the default)'
    )

    encryption_key: typing.Optional[str] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('encryption-key'),
        default=None,
        description='enable encryption and use the key for that'
    )

    extra_meta: typing.Optional[typing.Dict[str, typing.Any]] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('extra-meta'),
        default=None,
        description='some extra data to include to the archive'
    )

    archive: str = pydantic.Field(
        description='a target path to archive file'
    )

    backup_source: BackupSource = pydantic.Field(
        validation_alias=pydantic.AliasChoices('backup-source'),
        description='defines what we will backup'
    )


@register_bellboy_command()
class BellBoyBackupCommand(BellBoyCommandHandler):
    """ This command backs up data
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.BellBoyCommandHandler.command_name` method implementation
        """
        return "backup"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.BellBoyCommandHandler.command_model` method implementation
        """
        return BackupCommandModel

    def __read_command(self, command_str: str) -> IOGenerator:
        """ Start a command and return stdout

        :param str command_str: the command to start
        """
        with subprocess.Popen(shlex.split(command_str), stdout=subprocess.PIPE) as pipe:
            assert(pipe.stdout is not None)
            yield from IOThrottler.sync_reader(pipe.stdout)

    def __read_files_by_command(self, command_str: str) -> typing.Generator[str, None, None]:
        """ Start a command and yield file names to backup

        :param command_str: the command to start
        """
        with subprocess.Popen(shlex.split(command_str), stdout=subprocess.PIPE) as pipe:
            assert(pipe.stdout is not None)
            next_line = pipe.stdout.readline()
            while next_line != b'':
                yield next_line.decode().rstrip('\n')
                next_line = pipe.stdout.readline()

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.BellBoyCommandHandler.exec` method implementation
        """
        assert(isinstance(self._args, BackupCommandModel))

        archiver = BackupArchiveV1(
            hash_algorithms=self._args.hash_algorithms,
            throttling=self._args.throttling,
            compression=self._args.compression,
            cipher_name=self._args.cipher_name,
            encryption_key=self._args.encryption_key,
            extra_meta=self._args.extra_meta
        )

        with open(self._args.archive, 'wb') as archive_file:
            if self._args.backup_source.command is not None:
                await archiver.backup_io(self.__read_command(self._args.backup_source.command), archive_file)
            elif self._args.backup_source.files is not None:
                await archiver.backup_files(self._args.backup_source.files, archive_file)
            elif self._args.backup_source.files_command is not None:
                await archiver.backup_files(
                    self.__read_files_by_command(self._args.backup_source.files_command), archive_file
                )
            else:
                raise ValueError('Unknown backup source spotted!')

        with open(self._args.archive, 'rb') as archive_file:
            header = archiver.extract_header_meta(archive_file)
            tail = archiver.extract_tail_meta(archive_file)

        return LobbyKeyValueFeedbackResult(kv_result={
            "archiver_version": header.version,
            "archive_type": header.type.value,
            "created": header.created,
            "compression": header.compression.value,
            "cipher": header.cipher,
            "write_rate": tail.write_rate,
            "duration": tail.duration,
            "hashes": {x.algorithm.value: base64.b64encode(x.digest) for x in tail.hashes},
            "extra": header.extra,
        })


class ArchiveValidateCommandModel(pydantic.BaseModel):
    """ These settings define archive validation options
    """

    throttling: typing.Optional[int] = pydantic.Field(
        default=None,
        description='target data-rate (bytes per second) for archive reading'
    )

    archive: str = pydantic.Field(
        description='a target path to archive file'
    )


@register_bellboy_command()
class BellBoyArchiveValidateCommand(BellBoyCommandHandler):
    """ This command validate archive data
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.BellBoyCommandHandler.command_name` method implementation
        """
        return "backup-validate"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.BellBoyCommandHandler.command_model` method implementation
        """
        return ArchiveValidateCommandModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.BellBoyCommandHandler.exec` method implementation
        """
        assert(isinstance(self._args, ArchiveValidateCommandModel))

        with open(self._args.archive, 'rb') as archive_file:
            await BackupArchiveV1.validate_archive(archive_file, throttling=self._args.throttling)

        return LobbyStrFeedbackResult(str_result=f'The "{self._args.archive}" archive is consistent!')


class RestoreCommandModel(pydantic.BaseModel):
    """ These settings define unarchiving options
    """

    validate_hashes: pydantic_settings.CliImplicitFlag[bool] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('validate-hashes'),
        default=False,
        description='Whether to check hashes during restore or not'
    )

    throttling: typing.Optional[int] = pydantic.Field(
        default=None,
        description='target data-rate (bytes per second) for archive restoring'
    )

    encryption_key: typing.Optional[str] = pydantic.Field(
        validation_alias=pydantic.AliasChoices('encryption-key'),
        default=None,
        description='enable encryption and use the key for that'
    )

    archive: str = pydantic.Field(
        description='a target path to archive file'
    )

    restore_location: str = pydantic.Field(
        validation_alias=pydantic.AliasChoices('restore-location'),
        description='a target path to archive file'
    )


@register_bellboy_command()
class BellBoyRestoreCommand(BellBoyCommandHandler):
    """ These settings define unarchive options
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.BellBoyCommandHandler.command_name` method implementation
        """
        return "backup-restore"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.BellBoyCommandHandler.command_model` method implementation
        """
        return RestoreCommandModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.BellBoyCommandHandler.exec` method implementation
        """
        assert(isinstance(self._args, RestoreCommandModel))

        archive = self._args.archive
        restore_location = self._args.restore_location

        with open(restore_location, 'wb') as restore_file:
            with open(archive, 'rb') as archive_file:
                unarchiver = BackupArchiveV1.extract_data(
                    archive_file, self._args.encryption_key, validate_hashes=self._args.validate_hashes
                )
                await cag(IOThrottler.async_writer(unarchiver, restore_file, throttling=self._args.throttling))

        return LobbyStrFeedbackResult(
            str_result=f'The "{archive}" archive has been restored to the "{restore_location}" file'
        )



# -*- coding: utf-8 -*-

import gc
import pytest
import threading
import typing

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SignalsRegistry, CallbackRegistry

from pyknic.lib.signals.extra import CallbackWrapper
from pyknic.lib.signals.proto import Signal, SignalCallbackType, SignalSourceProto, SignalProxyProto
from pyknic.lib.signals.source import SignalSource
from pyknic.lib.tasks.threaded_task import ThreadedTask

from pyknic.lib.signals.proxy import SignalProxy, QueueProxy, QueueProxyStateError, QueueCallbackException


def test_exceptions() -> None:
    assert(issubclass(QueueProxyStateError, Exception) is True)
    assert(issubclass(QueueCallbackException, Exception) is True)


class TestSignalProxy:

    def test(self, signals_registry: 'SignalsRegistry') -> None:
        class Source(SignalSource):
            signal1 = Signal()

        signal_proxy = SignalProxy()
        source = Source()
        source.callback(Source.signal1, signal_proxy.proxy(signals_registry))

        source.emit(Source.signal1)
        assert(signals_registry.dump(True) == [
            (source, Source.signal1, None),
        ])

        signal_proxy.discard_proxy(signals_registry)
        source.emit(Source.signal1)
        assert(signals_registry.dump(True) == [])

    def test_default_wrapper(self, signals_registry: 'SignalsRegistry') -> None:
        class Source(SignalSource):
            signal1 = Signal()

        class CallbackCls:

            def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
                signals_registry(*args, **kwargs)

        signal_proxy = SignalProxy()
        source = Source()
        callback_obj = CallbackCls()
        source.callback(Source.signal1, signal_proxy.proxy(callback_obj))

        source.emit(Source.signal1)
        assert (signals_registry.dump(True) == [
            (source, Source.signal1, None),
        ])

        del callback_obj
        gc.collect()
        source.emit(Source.signal1)
        assert(signals_registry.dump(True) == [])  # since signal_proxy is collected

    def test_custom_wrapper(self, signals_registry: 'SignalsRegistry', callbacks_registry: 'CallbackRegistry') -> None:

        class Source(SignalSource):
            signal1 = Signal()

        class CustomCallbackWrapper(CallbackWrapper):

            def _pre_hook(
                self, callback: SignalCallbackType, source: SignalSourceProto, signal: Signal, value: typing.Any
            ) -> bool:
                callbacks_registry.callback('test_callback')()
                return True

        signal_proxy = SignalProxy(wrapper_factory=CustomCallbackWrapper)
        source = Source()
        source.callback(Source.signal1, signal_proxy.proxy(signals_registry))

        assert(callbacks_registry.calls('test_callback') == 0)
        source.emit(Source.signal1)
        assert(signals_registry.dump(True) == [
            (source, Source.signal1, None),
        ])
        assert(callbacks_registry.calls('test_callback') == 1)

        signal_proxy.discard_proxy(signals_registry)
        source.emit(Source.signal1)
        assert(signals_registry.dump(True) == [])
        assert(callbacks_registry.calls('test_callback') == 1)


class TestQueueProxy:

    def test_plain(self) -> None:
        queue_proxy = QueueProxy()
        threaded_task = ThreadedTask(queue_proxy)

        threaded_task.start()
        threaded_task.stop()
        threaded_task.join()

        assert(isinstance(queue_proxy, SignalProxyProto))

    def test_signals(self) -> None:
        queue_proxy = QueueProxy()
        threaded_task = ThreadedTask(queue_proxy)
        threaded_task.start()

        class Source(SignalSource):
            signal1 = Signal()
            signal2 = Signal(int)

        lock = threading.Lock()
        results = []

        def callback(source: SignalSourceProto, signal: Signal, value: typing.Any) -> None:
            nonlocal results, lock  # noqa: F824
            with lock:
                results.append((source, signal, value))

        source1 = Source()
        source2 = Source()

        source1.callback(Source.signal1, queue_proxy.proxy(callback))
        source1.callback(Source.signal2, queue_proxy.proxy(callback))
        source2.callback(Source.signal1, queue_proxy.proxy(callback))

        source1.emit(Source.signal1)
        source1.emit(Source.signal2, 0)
        source2.emit(Source.signal1)
        source2.emit(Source.signal2, 1)

        threaded_task.stop()
        threaded_task.join()

        assert(results == [
            (source1, Source.signal1, None),
            (source1, Source.signal2, 0),
            (source2, Source.signal1, None),
        ])

    def test_exec(self) -> None:
        queue_proxy = QueueProxy()
        threaded_task = ThreadedTask(queue_proxy)
        clbk_event = threading.Event()

        assert(queue_proxy.is_running() is False)

        threaded_task.start()

        def callback() -> None:
            nonlocal clbk_event  # noqa: F824
            clbk_event.set()

        queue_proxy.exec(callback)
        clbk_event.wait()
        assert(queue_proxy.is_running() is True)

        threaded_task.stop()
        threaded_task.join()
        assert(queue_proxy.is_running() is False)

    def test_exec_wait(self) -> None:
        queue_proxy = QueueProxy()
        callback_result = object()
        threaded_task = ThreadedTask(queue_proxy)
        threaded_task.start()

        def callback() -> object:
            nonlocal callback_result  # noqa: F824
            return callback_result

        assert(queue_proxy.exec(callback, blocking=True) is callback_result)

        threaded_task.stop()
        threaded_task.join()

    def test_exec_wait_exception(self) -> None:
        queue_proxy = QueueProxy()
        callback_exception = ValueError('!')
        threaded_task = ThreadedTask(queue_proxy)
        threaded_task.start()

        def callback() -> None:
            nonlocal callback_exception  # noqa: F824
            raise callback_exception

        with pytest.raises(QueueCallbackException):
            queue_proxy.exec(callback, blocking=True)

        try:
            queue_proxy.exec(callback, blocking=True)
        except QueueCallbackException as e:
            assert(e.__cause__ is callback_exception)

        threaded_task.stop()
        threaded_task.join()

    def test_state_exception(self) -> None:
        queue_proxy = QueueProxy()
        threaded_task = ThreadedTask(queue_proxy)

        with pytest.raises(QueueProxyStateError):
            queue_proxy.stop()

        with pytest.raises(QueueProxyStateError):
            queue_proxy.exec(lambda: None)

        threaded_task.start()

        with pytest.raises(QueueProxyStateError):
            queue_proxy.start()

        threaded_task.stop()
        threaded_task.join()

        with pytest.raises(QueueProxyStateError):
            queue_proxy.exec(lambda: None)

        with pytest.raises(QueueProxyStateError):
            queue_proxy.stop()

        with pytest.raises(QueueProxyStateError):
            queue_proxy.start()

    def test_is_inside(self) -> None:
        queue_proxy = QueueProxy()
        threaded_task = ThreadedTask(queue_proxy)
        threaded_task.start()

        def callback() -> object:
            nonlocal queue_proxy  # noqa: F824
            return queue_proxy.is_inside()

        assert(queue_proxy.exec(callback, blocking=True) is True)
        assert(callback() is False)

        threaded_task.stop()
        threaded_task.join()



# -*- coding: utf-8 -*-

import pytest
import time
import typing

from datetime import datetime, timezone

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import SignalsRegistry, SampleTasks

from pyknic.lib.signals.proto import SignalSourceProto
from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.proto import ScheduledTaskPostponePolicy, ScheduleRecordProto
from pyknic.lib.tasks.scheduler.queue import SchedulerQueue


scheduler_queue_test_case = (
    (
        {'policy': None, 'group': None},
    ),

    (
        {'policy': ScheduledTaskPostponePolicy.drop, 'group': None},
    ),

    (
        {'policy': None, 'group': 'group1'},
        {'policy': None, 'group': 'group2'},
    ),

    (
        {'policy': ScheduledTaskPostponePolicy.wait, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.wait, 'group': 'group2'},
        {'policy': ScheduledTaskPostponePolicy.wait, 'group': 'group1'},
    ),

    (
        {'policy': ScheduledTaskPostponePolicy.keep_last, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.keep_last, 'group': 'group2'},
        {'policy': ScheduledTaskPostponePolicy.keep_last, 'group': None},
        {'policy': None, 'group': None},
        {'policy': ScheduledTaskPostponePolicy.keep_last, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.keep_last, 'group': None},
    ),

    (
        {'policy': None, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.keep_first, 'group': 'group2'},
        {'policy': ScheduledTaskPostponePolicy.keep_first, 'group': None},
        {'policy': None, 'group': None},
        {'policy': None, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.keep_first, 'group': 'group1'},
        {'policy': ScheduledTaskPostponePolicy.keep_first, 'group': None},
    ),
)


class TestSchedulerPostponeQueue:

    def flush_records(self, queue: SchedulerQueue) -> typing.List[ScheduleRecordProto]:
        result = []
        next_record = queue.next_record()
        while next_record:
            result.append(next_record)
            next_record = queue.next_record()

        return result

    def test_plain(self, sample_tasks: 'SampleTasks') -> None:
        queue = SchedulerQueue()
        assert(isinstance(queue, SignalSourceProto) is True)

        record = sample_tasks.PlainRecord(PlainTask(lambda: None))
        queue.postpone(record)

        assert(self.flush_records(queue) == [record])

    @pytest.mark.parametrize(
        "records, result_indices", tuple(
            zip(
                scheduler_queue_test_case,
                (
                    (0, ),
                    tuple(),
                    (0, 1),
                    (0, 1, 2),
                    (1, 2, 3, 4, 5),
                    (0, 1, 2, 3, 6),
                )
            )
        )
    )
    def test_groups(
        self,
        sample_tasks: 'SampleTasks',
        records: typing.Tuple[typing.Dict[str, typing.Any]],
        result_indices: typing.Tuple[int]
    ) -> None:
        queue = SchedulerQueue()
        task = PlainTask(lambda: None)
        input_records = [
            sample_tasks.PlainRecord(task, postpone_policy=x['policy'], group_id=x['group']) for x in records
        ]
        result_records = [input_records[x] for x in result_indices]

        for i in input_records:
            queue.postpone(i)

        assert(self.flush_records(queue) == result_records)

    def test_ttl(self, sample_tasks: 'SampleTasks', signals_registry: 'SignalsRegistry') -> None:
        queue = SchedulerQueue()
        queue.callback(SchedulerQueue.task_expired, signals_registry)
        task = PlainTask(lambda: None)

        record1 = sample_tasks.PlainRecord(  # this one is kept
            task, ttl=(datetime.now(timezone.utc).timestamp() + 1000)
        )
        record2 = sample_tasks.PlainRecord(  # this one is dropped
            task, ttl=(datetime.now(timezone.utc).timestamp() - 10)
        )
        queue.postpone(record1)
        queue.postpone(record2)

        assert(self.flush_records(queue) == [record1])
        assert(signals_registry.dump(True) == [(queue, SchedulerQueue.task_expired, record2)])

    def test_ttl_next_record(self, sample_tasks: 'SampleTasks', signals_registry: 'SignalsRegistry') -> None:
        queue = SchedulerQueue()
        queue.callback(SchedulerQueue.task_expired, signals_registry)
        task = PlainTask(lambda: None)
        time_delta = 0.5

        record = sample_tasks.PlainRecord(task, ttl=(datetime.now(timezone.utc).timestamp() + time_delta))
        queue.postpone(record)

        time.sleep(time_delta * 2)
        assert(self.flush_records(queue) == [])
        assert(signals_registry.dump(True) == [(queue, SchedulerQueue.task_expired, record)])

    @pytest.mark.parametrize(
        "records, result_indices", tuple(
            zip(
                scheduler_queue_test_case,
                (
                    tuple(),
                    (0, ),
                    tuple(),
                    tuple(),
                    (0,),
                    (4, 5,),
                )
            )
        )
    )
    def test_dropped_signals(
        self,
        sample_tasks: 'SampleTasks',
        signals_registry: 'SignalsRegistry',
        records: typing.Tuple[typing.Dict[str, typing.Any]],
        result_indices: typing.Tuple[int]
    ) -> None:
        queue = SchedulerQueue()
        queue.callback(SchedulerQueue.task_dropped, signals_registry)

        task = PlainTask(lambda: None)
        input_records = [
            sample_tasks.PlainRecord(task, postpone_policy=x['policy'], group_id=x['group']) for x in records
        ]
        dropped_signals = [
            (queue, SchedulerQueue.task_dropped, input_records[x]) for x in result_indices
        ]

        for i in input_records:
            queue.postpone(i)

        assert(signals_registry.dump(True) == dropped_signals)

    @pytest.mark.parametrize(
        "records, result_indices", tuple(
            zip(
                scheduler_queue_test_case,
                (
                    (0, ),
                    tuple(),
                    (0, 1),
                    (0, 1, 2),
                    (0, 1, 2, 3, 4, 5),
                    (0, 1, 2, 3, 4, 6),
                )
            )
        )
    )
    def test_postponed_signals(
        self,
        sample_tasks: 'SampleTasks',
        signals_registry: 'SignalsRegistry',
        records: typing.Tuple[typing.Dict[str, typing.Any]],
        result_indices: typing.Tuple[int]
    ) -> None:
        queue = SchedulerQueue()
        queue.callback(SchedulerQueue.task_postponed, signals_registry)

        task = PlainTask(lambda: None)
        input_records = [
            sample_tasks.PlainRecord(task, postpone_policy=x['policy'], group_id=x['group']) for x in records
        ]
        postponed_signals = [
            (queue, SchedulerQueue.task_postponed, input_records[x]) for x in result_indices
        ]

        for i in input_records:
            queue.postpone(i)

        assert(signals_registry.dump(True) == postponed_signals)

    def test_filtered_next_record(self, sample_tasks: 'SampleTasks') -> None:
        queue = SchedulerQueue()
        task = PlainTask(lambda: None)

        record1 = sample_tasks.PlainRecord(task, group_id='group1')
        record2 = sample_tasks.PlainRecord(task, group_id='group2')
        queue.postpone(record1)
        queue.postpone(record2)

        assert(queue.next_record(lambda x: x.group_id() == 'group2') is record2)
        assert(queue.next_record(lambda x: x.group_id() == 'group2') is None)
        assert(queue.next_record() is record1)
        assert(queue.next_record() is None)

    def test_len(self, sample_tasks: 'SampleTasks') -> None:
        queue = SchedulerQueue()
        task = PlainTask(lambda: None)
        assert(len(queue) == 0)

        record1 = sample_tasks.PlainRecord(task, group_id='group1')
        record2 = sample_tasks.PlainRecord(task, group_id='group2')
        queue.postpone(record1)
        queue.postpone(record2)
        assert(len(queue) == 2)

        queue.next_record()
        assert(len(queue) == 1)
        queue.next_record()
        assert(len(queue) == 0)



# -*- coding: utf-8 -*-
# pyknic/lib/property.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing


DescriptorType = typing.TypeVar('DescriptorType')


class TypedDescriptor(typing.Generic[DescriptorType]):
    """This is a @property-a-like class that helps to restrict variable type and possible stored values."""

    def __init__(
        self,
        t: typing.Type[DescriptorType],
        default_value: typing.Optional[DescriptorType] = None,
        value_verifier: typing.Optional[typing.Callable[[typing.Optional[DescriptorType]], bool]] = None
    ) -> None:
        """Create a new descriptor -- https://docs.python.org/3/howto/descriptor.html

        :param t: attribute type
        :param default_value: attribute initial value (None will be used by default)
        :param value_verifier: function that checks attribute values (must return True on valid values
        and False otherwise)
        """
        self.__type = t
        self.__name: typing.Optional[str] = None
        self.__default_value: typing.Optional[DescriptorType] = default_value
        self.__value_verifier = value_verifier

        if self.__default_value is not None:
            self.__check_var(self.__default_value)

    def __check_var(self, value: typing.Optional[DescriptorType]) -> None:
        """Check value and raise exception if it is not valid.

        :param value: value to check
        """
        if value is not None and not isinstance(value, self.__type):
            raise TypeError(f"Value's type is invalid. It must be of type {self.__type.__class__.__name__}")

        if self.__value_verifier and not self.__value_verifier(value):
            raise ValueError('Value does not match specified verifier')

    def __get__(
        self, instance: typing.Any, owner: typing.Optional[typing.Any] = None
    ) -> typing.Optional[DescriptorType]:
        """Return attribute value."""
        if not self.__name:
            raise TypeError('Attribute name was not set')
        return typing.cast(DescriptorType, instance.__dict__.get(self.__name, self.__default_value))

    def __set__(self, instance: typing.Any, value: typing.Optional[DescriptorType]) -> typing.Any:
        """Set attribute value."""
        if not self.__name:
            raise TypeError('Attribute name was not set')
        self.__check_var(value)
        instance.__dict__[self.__name] = value

    def __set_name__(self, owner: typing.Any, name: str) -> None:
        """Receive attribute name

        https://docs.python.org/3/reference/datamodel.html#object.__set_name__
        """
        self.__name = name



# -*- coding: utf-8 -*-

import typing

from pyknic.lib.tasks.plain_task import PlainTask
from pyknic.lib.tasks.proto import TaskProto, TaskResult

if typing.TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from conftest import CallbackRegistry, SignalsRegistry


class TestPlainTask:

    def test(self, callbacks_registry: 'CallbackRegistry') -> None:
        plain_task = PlainTask(callbacks_registry.callback())
        assert(isinstance(plain_task, TaskProto) is True)
        plain_task.start()
        assert(callbacks_registry.total_calls() == 1)

    def test_success_result(
        self, callbacks_registry: 'CallbackRegistry',  signals_registry: 'SignalsRegistry'
    ) -> None:
        callback_result = object()
        plain_task = PlainTask(callbacks_registry.callback(callback_result=callback_result))
        plain_task.callback(plain_task.task_started, signals_registry)
        plain_task.callback(plain_task.task_completed, signals_registry)

        assert(signals_registry.dump(True) == [])
        plain_task.start()
        assert(signals_registry.dump(True) == [
            (plain_task, TaskProto.task_started, None),
            (plain_task, TaskProto.task_completed, TaskResult(callback_result)),
        ])

    def test_exception_result(
        self, callbacks_registry: 'CallbackRegistry', signals_registry: 'SignalsRegistry'
    ) -> None:
        callback_exception = ValueError('!')

        def exc() -> None:
            raise callback_exception

        plain_task = PlainTask(callbacks_registry.callback(callback_result=exc))
        plain_task.callback(plain_task.task_started, signals_registry)
        plain_task.callback(plain_task.task_completed, signals_registry)

        assert(signals_registry.dump(True) == [])
        plain_task.start()
        assert(signals_registry.dump(True) == [
            (plain_task, TaskProto.task_started, None),
            (plain_task, TaskProto.task_completed, TaskResult(exception=callback_exception)),
        ])



# -*- coding: utf-8 -*-
# pyknic/lib/thread.py
#
# Copyright (C) 2017-2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import functools
import threading
import types
import typing

import decorator
from pyknic.lib.typing import GenericFunc, P, R


class CriticalSectionError(Exception):
    """ An exception that may be raised if a lock for a critical section could not be acquired
    """
    pass


def acquire_lock(lock: threading.Lock, timeout: typing.Union[int, float, None] = None) -> bool:
    """ This function overload the original :meth:`.Lock.acquire` method that has two arguments ('blocking'
    and 'timeout') with the one argument only - 'timeout'. This argument may be set to None now

    :param lock: a lock that should be gained

    :param timeout: timeout that this function should wait in order to gain a lock. If the "timeout" is not
    specified (is None) then wait forever in a blocking mode till a lock is gained, if the "timeout" is a
    positive value this is a period of time in seconds that this function will wait for a lock in a blocking mode.
    If a non-positive value is set then this function will return immediately even if lock was not gained
    :type timeout: int | float | None

    :return: whether a specified lock is gained or not
    """
    if timeout is None:
        blocking = True
        timeout = -1
    elif timeout <= 0:
        blocking = False
        timeout = -1
    else:
        blocking = True
    return lock.acquire(blocking=blocking, timeout=timeout)


def critical_section_dynamic_lock(
    lock_fn: typing.Callable[..., threading.Lock],
    timeout_fn: typing.Optional[typing.Callable[..., typing.Union[int, float, None]]] = None
) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
    """ Protect a function with a lock, that was get from the specified function. If a lock can not be acquired, then
    no function call will be made

    :param lock_fn: callable that returns a lock, with which a function may be protected
    :param timeout_fn: callable that returns a timeout that is used the same way as the 'timeout' in
    the :func:`.acquire_lock` function

    :return: decorator with which a target function may be protected
    """

    def first_level_decorator(
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[..., typing.Any]:
        def second_level_decorator(
            original_function: typing.Callable[..., typing.Any], *args: typing.Any, **kwargs: typing.Any
        ) -> typing.Any:
            lock = lock_fn(*args, **kwargs)
            timeout = timeout_fn(*args, **kwargs) if timeout_fn else None
            if acquire_lock(lock, timeout=timeout) is True:
                try:
                    result = original_function(*args, **kwargs)
                    return result
                finally:
                    lock.release()

            raise CriticalSectionError('Unable to lock a critical section')

        return decorator.decorator(  # type: ignore[no-any-return, no-untyped-call]
            second_level_decorator
        )(decorated_function)
    return first_level_decorator


def critical_section_lock(
    lock: threading.Lock, timeout: typing.Union[int, float, None] = None
) -> typing.Callable[[GenericFunc[P, R]], GenericFunc[P, R]]:
    """ A wrapper for :func:`.critical_section_dynamic_lock` function call, but with a static lock and timeout objects
    instead of functions

    :param lock: lock with which a function will be protected
    :param timeout: timeout for acquiring lock

    :return: decorator with which a target function may be protected
    """

    def lock_getter(*args: typing.Any, **kwargs: typing.Any) -> threading.Lock:
        return lock

    def timeout_getter(*args: typing.Any, **kwargs: typing.Any) -> typing.Union[int, float, None]:
        return timeout

    return critical_section_dynamic_lock(lock_fn=lock_getter, timeout_fn=timeout_getter)


class CriticalResource:
    """ Class for simplifying thread safety for a shared object. Each class instance holds its own lock object
    that :meth:`.CriticalResource.critical_section` method will be used to protect bounded methods.
    Bounded methods are allowed to be protected only. Static methods or class methods are unsupported
    """

    class LockFreeContext:
        """ This class may be used as a context that gets a monopoly access to a critical resource
        (an :class:`.CriticalResource` object) and all the methods of that resource won't try to access already
        gained lock and won't cause a deadlock.

        :note: this object doesn't protect itself from a race condition. So this object usage must be limited to a
        single thread only
        """

        def __init__(self, critical_resource: 'CriticalResource', timeout: typing.Union[int, float, None] = None):
            """ Create a new context object

            :param critical_resource: resource that this context should access and protect
            :param timeout: the same as 'timeout' in the :func:`.acquire_lock` function
            """

            self.__resource = critical_resource
            self.__timeout = timeout
            self.__unlock = False

        def __getattr__(self, item: str) -> typing.Any:
            """ "Proxy" access to an original resource. If the result is protected by a lock with
            :meth:`.CriticalResource.critical_section` decorator then an original (lock free) method is returned

            :raise CriticalSectionError: if a resource was not locked before this access
            """
            if not self.__unlock:
                raise CriticalSectionError(
                    'A resource access error. A lock to the resource must be gotten at first!'
                )

            r_item = object.__getattribute__(self.__resource, item)

            try:
                return functools.partial(r_item.__pyknic_lock_free_fn__, r_item.__self__)
            except AttributeError:
                pass
            return r_item

        def __enter__(self) -> typing.Self:
            """ Enter this context and lock a resource

            :raise CriticalSectionError: if a resource was not locked during a specified timeout
            """
            if not acquire_lock(self.__resource.thread_lock(), self.__timeout):
                raise CriticalSectionError('Unable to lock a critical section')
            self.__unlock = True
            return self

        def __exit__(
            self,
            exc_type: typing.Optional[typing.Type[BaseException]],
            exc_val: typing.Optional[BaseException],
            exc_tb: typing.Optional[types.TracebackType]
        ) -> None:
            """ Exit a context and release a lock
            """
            if self.__unlock is True:
                self.__unlock = False
                self.__resource.thread_lock().release()

        def __del__(self) -> None:
            """ If a resource is locked still then unlock it
            """
            if self.__unlock is True:
                self.__unlock = False
                self.__resource.thread_lock().release()

    def __init__(self, timeout: typing.Union[int, float, None] = None) -> None:
        """ Create a lock

        :param timeout: the same as 'timeout' in the :func:`.acquire_lock` function
        """
        self.__lock = threading.Lock()
        self.__timeout = timeout

    def thread_lock(self) -> threading.Lock:
        """ Return a lock with which a bounded methods may be protected
        """
        return self.__lock

    def lock_timeout(self) -> typing.Union[int, float, None]:
        return self.__timeout

    def critical_context(self, timeout: typing.Union[int, float, None] = None) -> 'LockFreeContext':
        """ Create a lock free context for this object

        :param timeout: timeout for acquiring lock. the same as 'timeout' in
        the :meth:`.CriticalResource.LockFreeContext.__init__` method
        """
        return CriticalResource.LockFreeContext(self, timeout=timeout)

    @staticmethod
    def critical_section(
        decorated_function: typing.Callable[..., typing.Any]
    ) -> typing.Callable[..., typing.Any]:
        """ Decorate a method with a lock protection (class methods and static methods are unsupported)
        """

        def lock_fn(self: CriticalResource, *args: typing.Any, **kwargs: typing.Any) -> threading.Lock:
            if isinstance(self, CriticalResource) is False:
                raise TypeError(
                    'Invalid object type. It must be inherited from CriticalResource class and'
                    ' decorated method must be bounded'
                )
            return self.thread_lock()

        def timeout_fn(
            self: CriticalResource, *args: typing.Any, **kwargs: typing.Any
        ) -> typing.Union[int, float, None]:
            if isinstance(self, CriticalResource) is False:
                raise TypeError(
                    'Invalid object type. It must be inherited from CriticalResource class and'
                    ' decorated method must be bounded'
                )
            return self.lock_timeout()

        decorated_function.__pyknic_lock_free_fn__ = decorated_function  # type: ignore[attr-defined] # we are force
        # the attribute to be
        return critical_section_dynamic_lock(lock_fn=lock_fn, timeout_fn=timeout_fn)(decorated_function)



# -*- coding: utf-8 -*-

import asyncio
import pytest
import threading
import typing

from fixture_helpers import BaseFixture, pyknic_fixture


class EventLoopDescriptor:

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self.__aio_fixtures: typing.Set[typing.Awaitable[typing.Any]] = set()

    def add(self, task: typing.Awaitable[typing.Any]) -> None:
        self.__aio_fixtures.add(task)

    def wait(self, task: typing.Awaitable[typing.Any]) -> None:
        self.__aio_fixtures.remove(task)
        self.loop.run_until_complete(task)

    async def __wait_all(self) -> None:
        for t in self.__aio_fixtures:
            await t
        self.__aio_fixtures.clear()

    def close(self) -> None:
        self.loop.run_until_complete(self.__wait_all())
        self.loop.close()


class EventLoop(BaseFixture):

    __loop_lock__ = threading.Lock()
    __loops__: typing.Dict[threading.Thread, EventLoopDescriptor] = dict()

    @staticmethod
    def loop(*, init_loop: bool = False) -> EventLoopDescriptor:
        with EventLoop.__loop_lock__:
            if init_loop:
                assert(threading.current_thread() not in EventLoop.__loops__)
                result_loop = EventLoopDescriptor(asyncio.new_event_loop())
                EventLoop.__loops__[threading.current_thread()] = result_loop
                return result_loop
            else:
                assert(threading.current_thread() in EventLoop.__loops__)
                return EventLoop.__loops__[threading.current_thread()]

    @staticmethod
    def close_loop() -> None:
        with EventLoop.__loop_lock__:
            assert(threading.current_thread() in EventLoop.__loops__)
            loop_descriptor = EventLoop.__loops__.pop(threading.current_thread())
        loop_descriptor.close()

    @classmethod
    def start(cls) -> typing.Any:
        return EventLoop.loop(init_loop=True)

    @classmethod
    def finalize(cls, start_result: typing.Any) -> None:
        EventLoop.close_loop()


def _event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from pyknic_fixture(EventLoop)


@pytest.fixture
def event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from _event_loop()


@pytest.fixture(scope='class')
def class_event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from _event_loop()


@pytest.fixture(scope='module')
def module_event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from _event_loop()


@pytest.fixture(scope='package')
def package_event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from _event_loop()


@pytest.fixture(scope='session')
def session_event_loop() -> typing.Generator[asyncio.AbstractEventLoop, None, None]:
    yield from _event_loop()



# -*- coding: utf-8 -*-
# pyknic/lib/crypto/padding.py
#
# Copyright (C) 2018-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.io import IOGenerator, IOProducer
from pyknic.lib.crypto.proto import BlockPaddingProto
from pyknic.lib.verify import verify_value
from pyknic.lib.crypto.random import random_int
from pyknic.lib.io.aligner import Aligner


class SimplePadding(BlockPaddingProto):
    """ Class that pads given data with specified ASCII character
    """

    @verify_value(padding=lambda x: x is None or len(x) == 1)
    def __init__(self, padding: typing.Optional[bytes] = None, always_pad: bool = False):
        """ Create new padding class

        :param padding: integer code of ASCII character
        :param always_pad: whether a data should be always padded
        """
        BlockPaddingProto.__init__(self)

        if padding is None:
            padding = bytes([0])
        self.__padding_symbol = padding
        self.__always_pad = always_pad

        self.__processed_bytes = 0

    def padding_symbol(self) -> bytes:
        """ Return character with witch data is padded

        :return: bytes
        """
        return self.__padding_symbol

    def always_pad(self) -> bool:
        """Return True if padding will always be made event it is aligned already
        """
        return self.__always_pad

    @verify_value(block_size=lambda x: x > 0)
    def pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.pad` method implementation
        """

        if self.__processed_bytes != 0:
            raise RuntimeError('The "pad" or the "undo_pad" method has already been called')

        for i in data:
            self.__processed_bytes += len(i)
            yield i

        pad_length = block_size - (self.__processed_bytes % block_size)
        if pad_length == block_size and not self.__always_pad:
            pad_length = 0

        yield self.padding_symbol() * pad_length

    @verify_value(block_size=lambda x: x > 0)
    def undo_pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.undo_pad` method implementation
        """

        if self.__processed_bytes != 0:
            raise RuntimeError('The "pad" or the "undo_pad" method has already been called')

        aligner = Aligner(block_size, strict_mode=True)

        prev_block, next_block = None, None
        for aligner_block in aligner.iterate_data(data):
            if prev_block is not None:
                yield prev_block

            prev_block, next_block = next_block, aligner_block

        if prev_block is not None:
            yield prev_block

        if next_block is None:
            raise RuntimeError('There is no data. It seams that this data was not padded')

        if self.__always_pad and not next_block.endswith(self.padding_symbol()):
            raise RuntimeError('Invalid last symbol. It seams that this data was not padded')

        yield next_block.rstrip(self.padding_symbol())


class ZeroPadding(SimplePadding):
    """ Zero padding implementation (just alias for SimplePadding object)

    see also: https://en.wikipedia.org/wiki/Padding_(cryptography)#Zero_padding
    """

    def __init__(self) -> None:
        """ Create new padding object
        """
        SimplePadding.__init__(self, always_pad=True)


class ShiftPadding(SimplePadding):
    """ Same as :class:`.SimplePadding` class, but also randomly shifts original data.
    """

    def __init__(self, padding: typing.Optional[bytes] = None) -> None:
        """Create new padding object

        :param padding: same as the "padding" parameter in the :meth:`.SimplePadding.__init__` method
        """
        SimplePadding.__init__(self, padding=padding, always_pad=True)

    @verify_value(block_size=lambda x: x > 0)
    def pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.pad` method implementation
        """

        def shifted_data(original_data: IOProducer) -> IOGenerator:

            shift_length = random_int(block_size)
            if shift_length == 0:
                shift_length = 1

            yield self.padding_symbol() * shift_length

            for i in original_data:
                yield i

        return SimplePadding.pad(self, shifted_data(data), block_size)

    @verify_value(block_size=lambda x: x > 0)
    def undo_pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.undo_pad` method implementation
        """

        first_chunk: typing.Optional[bytes] = b''
        for i in SimplePadding.undo_pad(self, data, block_size):
            if first_chunk is None:
                yield i
                continue

            first_chunk += i
            first_chunk = first_chunk.lstrip(self.padding_symbol())
            if len(first_chunk):
                yield first_chunk
                first_chunk = None


class PKCS7Padding(BlockPaddingProto):
    """ PKCS7 Padding implementation

    see also: https://en.wikipedia.org/wiki/Padding_(cryptography)#PKCS7
    """

    def __init__(self) -> None:
        """ Create new padding class
        """
        BlockPaddingProto.__init__(self)
        self.__processed_bytes = 0

    @verify_value(block_size=lambda x: x > 0)
    def pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.pad` method implementation
        """

        if block_size >= 256:
            raise RuntimeError('The PKCS7 padding method is not supported for block size >= 256')

        if self.__processed_bytes != 0:
            raise RuntimeError('The "pad" or the "undo_pad" method has already been called')

        for i in data:
            self.__processed_bytes += len(i)
            yield i

        pad_length = block_size - (self.__processed_bytes % block_size)
        yield bytes([pad_length] * pad_length)

    @verify_value(block_size=lambda x: x > 0)
    def undo_pad(self, data: IOProducer, block_size: int) -> IOGenerator:
        """ :meth:`.BlockPaddingProto.undo_pad` method implementation
        """

        if block_size >= 256:
            raise RuntimeError('The PKCS7 padding method is not supported for block size >= 256')

        if self.__processed_bytes != 0:
            raise RuntimeError('The "pad" or the "undo_pad" method has already been called')

        aligner = Aligner(block_size, strict_mode=True)

        prev_block, next_block = None, None
        for aligner_block in aligner.iterate_data(data):
            if prev_block is not None:
                yield prev_block

            prev_block, next_block = next_block, aligner_block

        if prev_block is not None:
            yield prev_block

        if next_block is None:
            raise RuntimeError('There is no data. It seams that this data was not padded')

        padded = next_block[-1]
        if padded == 0 or padded > block_size:
            raise RuntimeError('Invalid last byte. It seams that this data was not padded with PKCS7')

        yield next_block[:-padded]



# -*- coding: utf-8 -*-
# pyknic/tasks/fastapi/lobby.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import typing
import secrets

import fastapi
import fastapi.encoders
import fastapi.responses
import fastapi.security

from pyknic.lib.config import Config
from pyknic.lib.registry import register_api
from pyknic.lib.fastapi.apps_registry import __default_fastapi_apps_registry__
from pyknic.lib.fastapi.models.lobby import LobbyFingerprintModel, LobbyCommandRequest
from pyknic.lib.gettext import GetTextWrapper
from pyknic.lib.fastapi.base import BaseFastAPIApp
from pyknic.lib.fastapi.lobby import LobbyCommandError, __default_lobby_commands_registry__
from pyknic.lib.fastapi.lobby_fingerprint import LobbyFingerprint
from pyknic.lib.fastapi.headers import FastAPIHeaders
from pyknic.lib.log import Logger


@register_api(__default_fastapi_apps_registry__, "lobby")
class LobbyApp(BaseFastAPIApp):
    """This web-app executes commands on remote server."""

    __default_modules__ = {
        'pyknic.lib.integrated_commands.ping',
        'pyknic.lib.integrated_commands.resources',
    }

    def __init__(self, config: Config, translations: GetTextWrapper):
        """Create a new web-app.

        :param config: config with which this app should be initialized
        :param translations: gettext access
        """

        BaseFastAPIApp.__init__(self, config, translations)

        self.__lobby_registry = __default_lobby_commands_registry__
        self.__translations = translations

        # TODO: make fingerprint survive restarts
        self.__fingerprint = LobbyFingerprint.generate_fingerprint()

    @classmethod
    def create_app(cls, fastapi_app: fastapi.FastAPI, config: Config, translations: GetTextWrapper) -> typing.Any:
        """ The :meth:`.FastAPIAppProto.create_app` method implementation
        """

        app = cls(config, translations)
        if not cls.secret_token(config):
            raise RuntimeError('Secret token is required')
        app.load_lobby_modules()

        fastapi_app.post(
            cls.lobby_main_path(config),
            status_code=200
        )(app.lobby_command)

        fastapi_app.get(
            cls.lobby_fingerprint_path(config),
            status_code=200,
            response_model=LobbyFingerprintModel
        )(app.fingerprint)

        return app

    def load_lobby_modules(self) -> None:
        """Import required modules so theirs commands may be used within this web-app."""
        modules_enabled = {str(x) for x in self.config()["pyknic"]["fastapi"]["lobby"]["modules"].iterate_list()}
        modules_enabled.update(self.__default_modules__)

        modules_list = list(modules_enabled)
        modules_list.sort()

        for module_name in modules_list:
            Logger.info(f'Import module "{module_name}" so commands will be loaded')
            importlib.import_module(module_name)

    def __sign_result(self, json_data: str) -> fastapi.Response:
        """Sign request result.

        :param json_data: API result to sign
        """
        signature = self.__fingerprint.sign(json_data.encode(), encode_base64=True)

        response_headers = dict()
        response_headers[FastAPIHeaders.fingerprint.value] = signature.decode('ascii')

        return fastapi.Response(content=json_data, media_type="application/json", headers=response_headers)

    async def fingerprint(self, request: fastapi.Request) -> LobbyFingerprintModel:
        return LobbyFingerprintModel(fingerprint=str(self.__fingerprint))

    async def login_with_bearer(self, request: fastapi.Request) -> None:
        """Authenticate request with bearer token.

        :param request: request with bearer token
        """
        # TODO: make it more secure
        #
        # And consider this:
        #
        # JWT and FastAPI:
        #   - https://fastapi.tiangolo.com/tutorial/security/oauth2-jwt/#handle-jwt-tokens
        # JWT known names:
        #   - https://pyjwt.readthedocs.io/en/stable/usage.html#registered-claim-names
        #   - https://en.wikipedia.org/wiki/JSON_Web_Token#Standard_fields
        #
        # JWT decode security warnings:
        #   - https://pyjwt.readthedocs.io/en/stable/api.html#jwt.decode

        http_bearer = fastapi.security.HTTPBearer()
        http_creds = await http_bearer(request)

        if http_creds:
            encoded_bearer = http_creds.credentials
            secret_token = self.secret_token(self.config())
            assert(secret_token)

            if secrets.compare_digest(encoded_bearer, secret_token):
                # note: compare_digest do comparison in a constant time
                return

        raise fastapi.HTTPException(
            status_code=fastapi.status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": 'Bearer realm="Token Required"'}
        )

    async def lobby_command(self, request: fastapi.Request, response: fastapi.Response) -> fastapi.Response:
        """Process command execution request.

        :param request: request with command request
        """
        await self.login_with_bearer(request)

        # TODO: check versions in client requests and the server's one. If they are differ, then result should have
        #   a warning about it. In future there may be plugins that has versions other than pyknic's

        try:
            if not await request.body():
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_400_BAD_REQUEST, detail='Please submit JSON within request'
                )

            data = await request.json()

            command_request = LobbyCommandRequest.model_validate(data)
            command_handler = self.__lobby_registry.get(command_request.name)
            command_args_class = command_handler.command_model()
            command_args = command_args_class.model_validate(command_request.args)

            command_result = await command_handler.exec(command_args)
            json_result = command_result.model_dump_json()
            return self.__sign_result(json_result)

        except LobbyCommandError as e:
            raise fastapi.HTTPException(status_code=fastapi.status.HTTP_400_BAD_REQUEST, detail=str(e))

    @classmethod
    def lobby_main_path(cls, config: Config) -> str:
        """Return path for command execution requests."""
        return str(config["pyknic"]["fastapi"]["lobby"]["main_url_path"])

    @classmethod
    def lobby_fingerprint_path(cls, config: Config) -> str:
        """Return path for fingerprint request."""
        return str(config["pyknic"]["fastapi"]["lobby"]["fingerprint_url_path"])

    @classmethod
    def secret_token(cls, config: Config) -> typing.Optional[str]:
        """Return secret token that is used for authentication."""
        token = config["pyknic"]["fastapi"]["lobby"]["secret_token"]
        return str(token) if not token.is_none() else None



# -*- coding: utf-8 -*-
# pyknic/lib/tasks/scheduler/plain_sources.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

from pyknic.lib.tasks.proto import ScheduleSourceProto, ScheduleRecordProto


class InstantTaskSource(ScheduleSourceProto):
    """ The simplest source that sends record immediately
    """

    def schedule_record(self, record: ScheduleRecordProto) -> None:
        """ Send a record

        :param record: a record to emit
        """
        self.emit(ScheduleSourceProto.task_scheduled, record)



# -*- coding: utf-8 -*-
# pyknic/lib/bellboy.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document it
# TODO: test it

import asyncio
import os
import pathlib
import sys
import typing

import pydantic
import pydantic_settings

from pyknic.lib.config import Config
from pyknic.lib.bellboy.console import BellboyConsole
from pyknic.lib.bellboy.models import CommonBellBoyCommandModel, FormattingMode
from pyknic.lib.bellboy.app import __default_bellboy_commands_registry__, BellBoyCommandHandler
from pyknic.tasks.log import LogTask

import pyknic.lib.integrated_commands  # noqa: F401  # force commands loading


# TODO: make the "rich" package optional dependency. But note! There is a direct call to the rich module inside
#  the "login" command

# TODO: think of loading external bellboy command. May be with the entry_points (think twice!)


class Bellboy:

    class BaseCommand(pydantic.BaseModel):
        bellboy: typing.Optional[CommonBellBoyCommandModel] = pydantic.Field(
            default=None,
            description='Utility configuration'
        )
        _command_handler: typing.Type[BellBoyCommandHandler]

    @classmethod
    def args_parser_cls(cls) -> pydantic_settings.BaseSettings:
        commands_annotations: typing.Dict[str, typing.Any] = dict()

        for command, handler in __default_bellboy_commands_registry__:
            assert(isinstance(command, str))
            handler_model: typing.Type[pydantic.BaseModel] = handler.command_model()

            class CustomModel(Bellboy.BaseCommand, handler_model):  # type: ignore[misc,valid-type]
                _command_handler = handler

            commands_annotations[command] = pydantic_settings.CliSubCommand[CustomModel]  # type: ignore[misc]

        class CommandsSpec:
            __annotations__ = commands_annotations

        class BellboySettings(
            CommandsSpec,
            pydantic_settings.BaseSettings,
            cli_parse_args=True,
            cli_prog_name='bellboy',
            cli_use_class_docs_for_groups=True
        ):
            """ Standalone pyknic tool and a client for the pyknic server
            """
            pass

        return BellboySettings()

    @classmethod
    def main(cls) -> None:
        config = Config()  # TODO: there is no usage at the moment =(

        with (pathlib.Path(__file__).parent / 'lib' / 'bellboy' / 'config.yaml').open() as f:
            config.merge_file(f)

        console = BellboyConsole()

        try:
            cmd = cls.args_parser_cls()
        except pydantic.ValidationError as e:
            console.error(str(e))
            sys.exit(-1)

        subcommand_obj = pydantic_settings.get_subcommand(cmd)
        assert(subcommand_obj is not None)

        assert(isinstance(subcommand_obj, Bellboy.BaseCommand))
        if subcommand_obj.bellboy is not None:
            if subcommand_obj.bellboy.config is not None:
                with pathlib.Path().open(subcommand_obj.bellboy.config):
                    config.merge_file(f)

        os.environ[LogTask.__env_var_name__] = "INFO"  # possible options are: "ERROR", "WARN", "INFO", "DEBUG"
        LogTask.setup_log()

        loop = asyncio.new_event_loop()
        prepared_command = subcommand_obj._command_handler.prepare_command(subcommand_obj)
        result = loop.run_until_complete(prepared_command.exec())

        json_mode = False
        if subcommand_obj.bellboy is not None:
            if subcommand_obj.bellboy.formatting is not None:
                if subcommand_obj.bellboy.formatting == FormattingMode.json:
                    json_mode = True
                else:
                    assert(subcommand_obj.bellboy.formatting == FormattingMode.rich)

        if json_mode:
            print(result.model_dump_json())
        else:
            console.process_result(result)


if __name__ == "__main__":
    Bellboy.main()



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/proto.py
#
# Copyright (C) 2017-2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
import typing
from abc import abstractmethod

from pyknic.lib.capability import CapabilitiesHolder, capability
from pyknic.lib.uri import URI

from pyknic.lib.verify import verify_value


class DirectoryNotEmptyError(Exception):
    """An exception is raised when there is a request to delete a directory which has inner files."""
    pass


class IOClientProto(CapabilitiesHolder):
    """ Base class for network clients. This class implements :class:`.WSchemeHandler` to handle connections
    encoded as URI and :class:`.WCapabilitiesHolder` to use capabilities as different client requests
    """

    # TODO: add basic async with usage, that will create, connect and disconnect client at the end
    # TODO: add signals that will notify about copying progress

    @classmethod
    @abstractmethod
    def create_client(cls, uri: URI) -> 'IOClientProto':
        raise NotImplementedError('This method is abstract')

    @capability
    async def connect(self) -> None:
        """Connect to a source specified in URI."""
        raise NotImplementedError('This method is abstract')

    @capability
    async def disconnect(self) -> None:
        """Disconnect from a source."""
        raise NotImplementedError('This method is abstract')

    @abstractmethod
    def uri(self) -> URI:
        """Return URI with which this client was created."""
        raise NotImplementedError('This method is abstract')

    @capability
    def current_directory(self) -> str:
        """Return current session directory"""
        raise NotImplementedError('This method is abstract')

    @capability
    async def change_directory(self, path: str) -> str:
        """Change current session directory to the specified one. If the path begins with directory separator
        then it may be treated as an absolute path.

        :return: new session directory
        """
        raise NotImplementedError('This method is abstract')

    @capability
    async def list_directory(self) -> typing.Tuple[str, ...]:
        """List current session directory."""
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def make_directory(self, directory_name: str) -> None:
        """Create directory. A directory is created in a current session directory. And a name must not
        contain a directory separator.

        :param directory_name: directory name to create
        """
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def remove_directory(self, directory_name: str) -> None:
        """Remove directory. A directory is removed from a current session directory. And a name must not
        contain a directory separator.

        :param directory_name: directory name to remove
        """
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    async def upload_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """Upload file. File will be uploaded to a current session directory. A name must not contain
        a directory separator

        :param remote_file_name: target file name
        :param local_file_obj: source object to upload
        """
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def remove_file(self, file_name: str) -> None:
        """Remove file. File will be removed from a current session directory. A name must not contain
        a directory separator

        :param file_name: file to remove
        """
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    async def receive_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        """Fetch/download a file. A remote file name must not contain a directory separator

        :param remote_file_name: file to fetch
        :param local_file_obj: a file where data should be stored
        """
        raise NotImplementedError('This method is abstract')

    @capability
    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    async def file_size(self, remote_file_name: str) -> int:
        """Return size of file in bytes.

        :param remote_file_name: file to check
        """
        raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# pyknic/lib/integrated_commands/resources.py
#
# Copyright (C) 2025-2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

# TODO: update tests!

import threading
import typing

import pydantic

from pyknic.lib.bellboy.app import BellBoyCommandHandler, register_bellboy_command, LobbyClient
from pyknic.lib.bellboy.models import GeneralBellBoyCommandModel
from pyknic.lib.fastapi.lobby import LobbyCommandHandler, register_lobby_command
from pyknic.lib.fastapi.models.base import NullableModel
from pyknic.lib.fastapi.models.lobby import LobbyCommandResult, LobbyKeyValueFeedbackResult, LobbyCommandRequest


@register_lobby_command()
class LobbyResourcesCommand(LobbyCommandHandler):
    """This is a server side API-"command" that takes no arguments and returns minor information about server's
    used resources
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.LobbyCommandHandler.command_name` method implementation
        """
        return "resources"

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.LobbyCommandHandler.command_model` method implementation
        """
        return NullableModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.LobbyCommandHandler.exec` method implementation
        """
        with open('/proc/self/statm') as mem_usage_fd:  # TODO: validate errors!
            statm = mem_usage_fd.readline()
            total, resident, shared = statm.split(' ')[:3]

        pythreads = threading.active_count()

        return LobbyKeyValueFeedbackResult(kv_result={
            "mem_total": int(total), "mem_resident": int(resident), "mem_shared": int(shared), "py_threads": pythreads
        })


@register_bellboy_command()
class BellBoyResourcesCommand(BellBoyCommandHandler):
    """This is a CLI "command" that returns minor information about server's used resources
    """

    @classmethod
    def command_name(cls) -> str:
        """ The :meth:`.BellBoyCommandHandler.command_name` method implementation
        """
        return LobbyResourcesCommand.command_name()

    @classmethod
    def command_model(cls) -> typing.Type[pydantic.BaseModel]:
        """ The :meth:`.BellBoyCommandHandler.command_model` method implementation
        """
        return GeneralBellBoyCommandModel

    async def exec(self) -> LobbyCommandResult:
        """ The :meth:`.BellBoyCommandHandler.exec` method implementation
        """
        assert(isinstance(self._args, GeneralBellBoyCommandModel))
        auth_data = self.auth_data(self._args.secret_backend, self._args.lobby_url)

        client = LobbyClient(self._args.lobby_url, auth_data.server_fingerprint, auth_data.token)
        return await client.command_request(
            LobbyCommandRequest(
                name=LobbyResourcesCommand.command_name(),
                args=NullableModel().model_dump()
            )
        )



# -*- coding: utf-8 -*-
# pyknic/lib/x_mansion.py
#
# Copyright (C) 2024 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import typing

from pyknic.lib.signals.source import SignalSourceMeta, SignalSource
from pyknic.lib.capability import CapabilitiesHolderMeta, CapabilitiesHolder


class CapabilitiesAndSignalsMeta(SignalSourceMeta, CapabilitiesHolderMeta):
    """ This metaclass is for classes that may send signals and may have capabilities
    """

    def __init__(cls, name: str, bases: typing.Tuple[type], namespace: typing.Dict[str, typing.Any]):
        """ Initialize class with this metaclass

        :param name: same as 'name' in :meth:`.WSignalSourceMeta.__init__` method
        :param bases: same as 'bases' in :meth:`.WSignalSourceMeta.__init__` method
        :param namespace: same as 'namespace' in :meth:`.WSignalSourceMeta.__init__` method
        """
        SignalSourceMeta.__init__(cls, name, bases, namespace)
        CapabilitiesHolderMeta.__init__(cls, name, bases, namespace)


class CapabilitiesAndSignals(SignalSource, CapabilitiesHolder, metaclass=CapabilitiesAndSignalsMeta):
    """ This is a base class for classes that may send signals and may have capabilities
    """
    pass
